#include "internal.h"

#include <stdlib.h>

static char initialised = 0;

static jfaud_channel midichan;
static jfaud_channel cdachan;
static jfaud_channel *pcmchan = NULL;

static char usemidi = 0, usecda = 0, usepcm = 0;

static struct jfaud_init_params *params;


jfauderr jfaudchan_init(void)
{
	jfauderr err;
	
	if (initialised) return jfauderr_alreadyinited;

	if ((err = jfaudbuffer_init()) != jfauderr_ok) return err;
	
	params = jfaud_getparams();
	if (!params) return jfauderr_genericerror;
	
	memset(&midichan, 0, sizeof(midichan));
	memset(&cdachan, 0, sizeof(cdachan));

	pcmchan = (jfaud_channel *)malloc(sizeof(jfaud_channel) * params->maxvoices);
	if (!pcmchan) return jfauderr_outofmemory;
	memset(pcmchan, 0, sizeof(jfaud_channel) * params->maxvoices);

	usemidi = params->mididevice != NULL;
	usecda = params->cdadevice != NULL;
	usepcm = params->wavedevice != NULL;
	// TODO: if midi device is Ken's adlib emulator, that requires pcm be enabled, so check
	
	if (usepcm) {
		// initialise openal
		err = jfaudopenal_init();
		if (err != jfauderr_ok) {
			params->call_logmsg("Failure initialising OpenAL");
			usepcm = 0;
		}
	}

	if (usemidi) {
		err = jfaudmidi_init();
		if (err != jfauderr_ok) {
			usemidi = 0;
		}
	}

	initialised = 1;

	return jfauderr_ok;
}

jfauderr jfaudchan_uninit(void)
{
	int i;
	if (!initialised) return jfauderr_notinited;

	// stop all channels
	if (pcmchan) {
		for (i=params->maxvoices-1; i>=0; i--)
			jfaudchan_release( (jfaudfiledata_pcm<<16)|i );
		free(pcmchan);
		pcmchan = NULL;
	}

	jfaudchan_release( (jfaudfiledata_midi<<16)|i );
	
	// stop all devices
	if (usepcm) {
		jfaudopenal_uninit();
	}

	if (usemidi) {
		jfaudmidi_uninit();
	}

	jfaudbuffer_uninit();

	return jfauderr_ok;
}

static jfauderr jfaudchan_updateone(jfaud_channel *ch)
{
	jfaud_buffer buf;
	jfauderr err;
	
	if (!ch->used) return jfauderr_genericerror;
	if (ch->age < 0xffffffff) ch->age++;
	if (!ch->data) return jfauderr_ok;	// no data source. This handles non-streamed sounds.
	if (ch->data->type == jfaudfile_unidentified) return jfauderr_unrecognised;

	if (filereaders[ ch->data->type ]->datatype == jfaudfiledata_pcm) {
		if (jfaudopenal_needsdata(ch->outhandle) != jfauderr_ok) return jfauderr_ok;	// no data needed

		do {
			err = filereaders[ ch->data->type ]->readpcmdata(ch->data, &buf, stream_samples);
			if (err == jfauderr_ok) {
				if (buf.numsamples == 0) {
					// end of file, so rewind and try again
					if (!ch->loop) return jfauderr_ok;
					filereaders[ ch->data->type ]->rewindpcmdata(ch->data);
					continue;
				}
				break;	// got a chunk of data
			} else {
				return err;
			}
		} while(1);

		if (simulateaudiolib)
			jfaudfx_audiolib(&buf);
		
		if ((ch->ident >> 32))	// double-speed hacked channel
			buf.samplerate <<= 1;

		err = jfaudopenal_preparebuffer(&buf);
		if (err != jfauderr_ok) {
			jfaudbuffer_free(&buf);
			return err;
		}
		
		err = jfaudopenal_queuedata(ch->outhandle, &buf);
		if (err != jfauderr_ok) {
			jfaudopenal_releasebuffer(&buf);
			return err;
		}
		
		// at this point the queue in openal owns the buffer

		//err = jfaudbuffer_free(&buf);
		//if (err != jfauderr_ok) return err;

		return jfauderr_ok;
	}

	if (filereaders[ ch->data->type ]->datatype == jfaudfiledata_midi) {
		return jfauderr_ok;
	}

	return jfauderr_unrecognised;
}

jfauderr jfaudchan_update(void)
{
	int i;

	if (!initialised) return jfauderr_notinited;
	
	if (usepcm) {
		jfaudopenal_update();
		for (i = params->maxvoices-1; i>=0; i--)
			jfaudchan_updateone(&pcmchan[i]);
		jfaudbuffer_update();
	}

	if (usemidi) {
		jfaudchan_updateone(&midichan);
	}

	return jfauderr_ok;
}

jfauderr jfaudchan_alloc(int *chan, jfaudident ident, int priority)
{
	// allocates a channel to play a managed sound

	int lowestchan, lowestprio, oldage, i;
	jfauderr err;
	jfaud_buffer *buf = NULL;

	if (!initialised) return jfauderr_notinited;
	if (!usepcm) return jfauderr_nowave;

	err = jfaudbuffer_fetch(ident, &buf);
	if (err != jfauderr_ok) return err;
	if (!buf) return jfauderr_notloaded;

	lowestchan = params->maxvoices;
	lowestprio = 0x7fffffff;
	oldage = 0;
	
	// find a free PCM channel, or knock one off if all are full
	for (i = params->maxvoices - 1; i>=0; i--) {
		if (!pcmchan[i].used) {
			lowestchan = i;
			lowestprio = 0;
			break;
		}
		if (pcmchan[i].priority != reserved_priority &&		// never turf out reserved chans
		    (pcmchan[i].priority < lowestprio ||		// lower prio channels are first to go
		     (pcmchan[i].priority == lowestprio && pcmchan[i].age > oldage)	// oldest chan of same prio dies first
		    )
		   ) {
			lowestchan = i;
			lowestprio = pcmchan[i].priority;
			oldage = pcmchan[i].age;
		}
	}

	if (priority < lowestprio && priority != reserved_priority) {
		jfaudbuffer_release(ident);
		return jfauderr_genericerror;
	}

	// got a channel. if something is playing already, kill it
	if (pcmchan[lowestchan].used) {
		debuglog(params->call_logmsg,jsprintf("channel: Kicking out PCM channel %d with priority %d, age %d",
					lowestchan,pcmchan[lowestchan].priority,pcmchan[lowestchan].age));
		jfaudopenal_releasesource(pcmchan[lowestchan].outhandle);
		if (pcmchan[lowestchan].data)
			jfaudfile_close(pcmchan[lowestchan].data);
	}

	memset(&pcmchan[lowestchan], 0, sizeof(jfaud_channel));

	err = jfaudopenal_allocsource(&pcmchan[lowestchan].outhandle, 0);
	if (err != jfauderr_ok) {
		jfaudbuffer_release(ident);
		return err;
	}

	pcmchan[lowestchan].used = 1;
	pcmchan[lowestchan].age = 0;
	pcmchan[lowestchan].data = NULL;
	pcmchan[lowestchan].ident = ident;
	pcmchan[lowestchan].stream = 0;
	pcmchan[lowestchan].priority = priority;

	err = jfaudopenal_queuedata(pcmchan[lowestchan].outhandle, buf);
	if (err != jfauderr_ok) {
		jfaudopenal_releasesource(pcmchan[lowestchan].outhandle);
		jfaudbuffer_release(ident);
		return err;
	}

	*chan = (jfaudfiledata_pcm<<16)|(lowestchan);
	return jfauderr_ok;
}

jfauderr jfaudchan_allocstream(int *chan, jfaud_file *data, int priority)
{
	// allocates a channel of the type suitable for the data source
	// this is where the openal, cda and midi devices converge
	
	// if a channel is kicked out because of it being lower priority,
	// the channel we return might collide with one thought to be in
	// use. Therefore, the channel needs to be tested against other
	// sources and the one overwritten should be silently closed.
	
	int lowestchan, lowestprio, oldage, i;
	jfauderr err;

	if (!initialised) return jfauderr_notinited;
	
	lowestchan = params->maxvoices;
	lowestprio = 0x7fffffff;
	oldage = 0;
	
	if (data->type == jfaudfile_unidentified) return jfauderr_unrecognised;

	if (filereaders[ data->type ]->datatype == jfaudfiledata_pcm) {
		if (!usepcm) return jfauderr_nowave;

		// find a free PCM channel, or knock one off if all are full
		for (i = params->maxvoices - 1; i>=0; i--) {
			if (!pcmchan[i].used) {
				lowestchan = i;
				lowestprio = 0;
				break;
			}
			if (pcmchan[i].priority != reserved_priority &&		// never turf out reserved chans
			    (pcmchan[i].priority < lowestprio ||		// lower prio channels are first to go
			     (pcmchan[i].priority == lowestprio && pcmchan[i].age > oldage)	// oldest chan of same prio dies first
			    )
			   ) {
				lowestchan = i;
				lowestprio = pcmchan[i].priority;
				oldage = pcmchan[i].age;
			}
		}

		if (priority < lowestprio && priority != reserved_priority)
			return jfauderr_genericerror;

		// got a channel. if something is playing already, kill it
		if (pcmchan[lowestchan].used) {
			debuglog(params->call_logmsg,jsprintf("channel: Kicking out PCM channel %d with priority %d, age %d",
						lowestchan,pcmchan[lowestchan].priority,pcmchan[lowestchan].age));
			jfaudopenal_releasesource(pcmchan[lowestchan].outhandle);
			if (pcmchan[lowestchan].data)
				jfaudfile_close(pcmchan[lowestchan].data);
		}

		memset(&pcmchan[lowestchan], 0, sizeof(jfaud_channel));

		/*
		if (filereaders[ data->type ]->getplaytime(data, &playtime) == jfauderr_ok) {
			debuglog(params->call_logmsg,jsprintf("channel: Sound ident %x plays %g sec, stream=%d",
						data->ident,playtime,stream));
			if (playtime >= stream_threshold) stream = 1;
		} else {
			debuglog(params->call_logmsg,jsprintf("channel: Could not get play time for sound ident %x",
						data->ident));
		}
		*/

		err = jfaudopenal_allocsource(&pcmchan[lowestchan].outhandle, 1);
		if (err != jfauderr_ok) return err;

		pcmchan[lowestchan].used = 1;
		pcmchan[lowestchan].data = data;
		pcmchan[lowestchan].stream = 1;
		pcmchan[lowestchan].priority = priority;

		err = jfaudchan_updateone(&pcmchan[lowestchan]);
		if (err != jfauderr_ok) {
			jfaudopenal_releasesource(pcmchan[lowestchan].outhandle);
			return err;
		}

		*chan = (jfaudfiledata_pcm<<16)|(lowestchan);
		return jfauderr_ok;
	}
	
	if (filereaders[ data->type ]->datatype == jfaudfiledata_midi) {
		if (!usemidi) return jfauderr_nomidi;
		// if using ken's adlib emulator and no pcm device, return jfauderr_nowave;

		if (midichan.used) {
			debuglog(params->call_logmsg,"channel: Kicking out MIDI channel");
			jfaudfile_close(midichan.data);
		}
		
		memset(&midichan, 0, sizeof(jfaud_channel));

		if (jfaudmidi_play((midiseq*)data->readerinfo, 0)) {
			return jfauderr_genericerror;
		}

		midichan.used = 1;
		midichan.data = data;
		midichan.stream = 0;
		midichan.priority = -1;

		*chan = (jfaudfiledata_midi<<16);
		return jfauderr_ok;
	}

	if (filereaders[ data->type ]->datatype == jfaudfiledata_cda) {
		if (!usecda) return jfauderr_nocda;
		return jfauderr_nocda;
	}
	
	return jfauderr_unrecognised;
}

jfauderr jfaudchan_release(int chan)
{
	int dachan;
	
	// frees a channel
	if (!initialised) return jfauderr_notinited;

	dachan = chan&0x0ffff;
	switch (chan>>8) {
		case jfaudfiledata_pcm:
			if (!pcmchan[dachan].used) return jfauderr_ok;	// already clear

			jfaudopenal_releasesource(pcmchan[dachan].outhandle);
			if (!pcmchan[dachan].stream) jfaudbuffer_release(pcmchan[dachan].ident);
			if (pcmchan[dachan].data) jfaudfile_close(pcmchan[dachan].data);

			memset(&pcmchan[dachan], 0, sizeof(jfaud_channel));

			return jfauderr_ok;

		case jfaudfiledata_midi:
			if (!midichan.used) return jfauderr_ok;

			jfaudmidi_stop();
			if (midichan.data) jfaudfile_close(midichan.data);

			memset(&midichan, 0, sizeof(jfaud_channel));

			return jfauderr_ok;

		default: break;
	}

	return jfauderr_unrecognised;
}

jfauderr jfaudchan_setprops(int chan, float volume, float pitch, char loop, char threed)
{
	int dachan;
	jfauderr err;
	
	if (!initialised) return jfauderr_notinited;

	dachan = chan&0x0ffff;
	switch (chan>>16) {
		case jfaudfiledata_pcm:
			if (!pcmchan[dachan].used) return jfauderr_ok;	// nobody home

			pcmchan[dachan].volume = volume;
			pcmchan[dachan].pitch = pitch;
			pcmchan[dachan].loop = loop;
			pcmchan[dachan].threed = threed;

			if (pcmchan[dachan].stream) loop = 0;

			err = jfaudopenal_setsourceprops(pcmchan[dachan].outhandle,
					volume, pitch, loop, threed);

			return err;

		case jfaudfiledata_midi:
			if (!midichan.used) return jfauderr_ok;

			jfaudmidi_setloop(loop);

			return jfauderr_ok;

		default: break;
	}

	return jfauderr_unrecognised;
}

jfauderr jfaudchan_setpos(int chan,
		float posx, float posy, float posz,
		float velx, float vely, float velz,
		char posrel)
{
	int dachan;
	jfauderr err;
	
	if (!initialised) return jfauderr_notinited;

	dachan = chan&0x0ffff;
	switch (chan>>16) {
		case jfaudfiledata_pcm:
			if (!pcmchan[dachan].used) return jfauderr_ok;	// nobody home

			err = jfaudopenal_setsourcepos(pcmchan[dachan].outhandle,
					posx, posy, posz,  velx, vely, velz,  posrel);

			pcmchan[dachan].posx = posx;
			pcmchan[dachan].posy = posy;
			pcmchan[dachan].posz = posz;
			pcmchan[dachan].velx = velx;
			pcmchan[dachan].vely = vely;
			pcmchan[dachan].velz = velz;
			pcmchan[dachan].posrel = posrel;

			return err;

		case jfaudfiledata_midi:
			return jfauderr_ok;

		default: break;
	}

	return jfauderr_unrecognised;
}

jfauderr jfaudchan_getplaymode(int chan, jfaudplaymode *mode)
{
	int dachan;
	jfauderr err;
	
	if (!initialised) return jfauderr_notinited;
	
	dachan = chan&0x0ffff;
	switch (chan>>16) {
		case jfaudfiledata_pcm:
			if (!pcmchan[dachan].used) return jfauderr_ok;	// nobody home

			*mode = pcmchan[dachan].playmode;
			err = jfaudopenal_getsourceplaymode(pcmchan[dachan].outhandle, mode);
			pcmchan[dachan].playmode = *mode;

			return err;

		case jfaudfiledata_midi:
			return jfaudmidi_getplaymode(mode);

		default: break;
	}

	return jfauderr_unrecognised;
}

jfauderr jfaudchan_setplaymode(int chan, jfaudplaymode mode)
{
	int dachan;
	jfauderr err;
	
	if (!initialised) return jfauderr_notinited;
	
	dachan = chan&0x0ffff;
	switch (chan>>16) {
		case jfaudfiledata_pcm:
			if (!pcmchan[dachan].used) return jfauderr_ok;	// nobody home

			pcmchan[dachan].playmode = mode;
			err = jfaudopenal_setsourceplaymode(pcmchan[dachan].outhandle, mode);

			return err;

		case jfaudfiledata_midi:
			if (!midichan.used) return jfauderr_ok;

			return jfaudmidi_setplaymode(mode);

		default: break;
	}

	return jfauderr_unrecognised;
}


