#include "internal.h"
#include <stddef.h>


char simulateaudiolib = 0;

static int musicchan = -1;

jfauderr jfaud_playmusic(const char *file, float volume, char loop, jfaudplaymode playmode)
{
	jfaud_file *fil;
	jfauderr err;

	if ((err = jfaudfile_open(file, &fil)) != jfauderr_ok) return err;

	if (fil->type == jfaudfile_unidentified) {
		jfaudfile_close(fil);
		return jfauderr_unrecognised;
	}
	
	if (musicchan >= 0) jfaudchan_release(musicchan);
	musicchan = -1;
	
	if ((err = jfaudchan_allocstream(&musicchan, fil, reserved_priority)) != jfauderr_ok) {
		jfaudfile_close(fil);
		return err;
	}

	jfaudchan_setprops(musicchan, volume, 1.0, loop, 0);
	jfaudchan_setpos(musicchan, 0.0,0.0,0.0, 0.0,0.0,0.0, 1);
	
	if (playmode == jfaud_playing)
		return jfaudchan_setplaymode(musicchan, playmode);
	
	return jfauderr_ok;
}

jfauderr jfaud_setmusicplaymode(jfaudplaymode playmode)
{
	if (playmode == jfaud_stopped) {
		if (musicchan < 0) return jfauderr_ok;

		jfaudchan_release(musicchan);
		musicchan = -1;

		return jfauderr_ok;
	} else if (playmode == jfaud_paused || playmode == jfaud_playing) {
		if (musicchan < 0) return jfauderr_notplaying;

		jfaudchan_setplaymode(musicchan, playmode);

		return jfauderr_ok;
	}

	return jfauderr_genericerror;
}

jfauderr jfaud_getmusicplaymode(jfaudplaymode *playmode)
{
	if (musicchan < 0) {
		*playmode = jfaud_stopped;
		return jfauderr_ok;
	}

	return jfaudchan_getplaymode(musicchan, playmode);
}



jfauderr jfaud_playsound(int *handle, const char *file,
		float volume, float pitch, char loop,
		float posx, float posy, float posz,
		float velx, float vely, float velz,
		char posrel, char threed, int priority, jfaudplaymode playmode)
{
	int h;
	jfaud_file *fil = NULL;
	jfauderr err;
	jfaudident ident;
	char stream = 0;
	unsigned int len;

	if (priority < 0) priority = 0;
	*handle = -1;

	ident = jfaud_hashfn(file);
	
	if (pitch > 2.0) {
		// Need to use the double-speed hack. This can't be disabled
		// once it's used.

		ident |= 0x100000000;
		pitch /= 2.0;
	}
	
	if (jfaudbuffer_isavailable(ident) == jfauderr_notloaded) {
		if ((err = jfaudfile_open(file, &fil)) != jfauderr_ok) return err;

		if (fil->type == jfaudfile_unidentified) {
			jfaudfile_close(fil);
			return jfauderr_unrecognised;
		} else if (filereaders[ fil->type ]->datatype != jfaudfiledata_pcm) {
			jfaudfile_close(fil);
			return jfauderr_wrongtype;
		}

		if ((err = jfaudfile_getpcmlength(fil, &len)) != jfauderr_ok) {
			jfaudfile_close(fil);
			return err;
		}

		if (len > stream_samples) stream = 1;
		else {
			jfaud_buffer buf;

			err = filereaders[ fil->type ]->readallpcmdata(fil, &buf);
			if (err == jfauderr_ok) {
				if (simulateaudiolib) jfaudfx_audiolib(&buf);

				if ((ident >> 32))	// double-speed hacked channel
					buf.samplerate <<= 1;

				err = jfaudopenal_preparebuffer(&buf);
				if (err != jfauderr_ok) {
					jfaudbuffer_free(&buf);
					jfaudfile_close(fil);
					return err;
				}

				// buffer is freed as a consequence of this

				err = jfaudbuffer_manage(ident, &buf);
				if (err != jfauderr_ok) {
					jfaudopenal_releasebuffer(&buf);
					jfaudfile_close(fil);
					return err;
				}

				// the buffer is now owned by the management system so we don't free it,
				// rather we just forget it exists.

				jfaudfile_close(fil);
			} else {
				jfaudfile_close(fil);
				return err;
			}
		}
	}

	if (stream) {
		if ((err = jfaudchan_allocstream(&h, fil, priority)) != jfauderr_ok) {
			jfaudfile_close(fil);
			return err;
		}

		// the channel owns the file now
	} else {
		if ((err = jfaudchan_alloc(&h, ident, priority)) != jfauderr_ok) {
			return err;
		}
	}

	*handle = h;

	jfaudchan_setprops(h, volume, pitch, loop, threed);
	jfaudchan_setpos(h, posx,posy,posz, velx,vely,velz, posrel);
	
	if (playmode == jfaud_playing)
		return jfaudchan_setplaymode(h, playmode);
	
	return jfauderr_ok;
}

jfauderr jfaud_movesound(int handle, float posx, float posy, float posz,
		float velx, float vely, float velz, char posrel)
{
	if (handle == musicchan) {
		// this happens if music is started and it stole the channel the sound
		// is playing on, and if that's the case, ignore the request
		return jfauderr_ok;
	}
	return jfaudchan_setpos(handle, posx,posy,posz, velx,vely,velz, posrel);
}

jfauderr jfaud_setpos(float posx, float posy, float posz,
		float velx, float vely, float velz,
		float atx, float aty, float atz,
		float upx, float upy, float upz)
{
	return jfaudopenal_setpos(posx,posy,posz,velx,vely,velz,atx,aty,atz,upx,upy,upz);
}

jfauderr jfaud_setsoundplaymode(int handle, jfaudplaymode playmode)
{
	if (handle == musicchan) {
		// see explanation in movesound
		return jfauderr_ok;
	}
	if (playmode == jfaud_stopped) {
		return jfaudchan_release(handle);
	} else if (playmode == jfaud_paused || playmode == jfaud_playing) {
		return jfaudchan_setplaymode(handle, playmode);
	}

	return jfauderr_genericerror;
}

jfauderr jfaud_getsoundplaymode(int handle, jfaudplaymode *playmode)
{
	if (handle == musicchan) {
		// see explanation in movesound
		*playmode = jfaud_stopped;
		return jfauderr_ok;
	}
	return jfaudchan_getplaymode(handle, playmode);
}

jfauderr jfaud_simulateaudiolib(int onf)
{
	simulateaudiolib = onf!=0;
	return jfauderr_ok;
}

