#include "internal.h"

#include <stdlib.h>

#if (WITH_VORBIS == 0)
static jfauderr init(void) { return jfauderr_unrecognised; }
static jfauderr uninit(void) { return jfauderr_unrecognised; }
static jfauderr testfile(jfaud_file *file) { return jfauderr_unrecognised; }
static jfauderr cleanup(jfaud_file *file, char *noclose) { return jfauderr_unrecognised; }
static jfauderr readpcmdata(jfaud_file *file, jfaud_buffer *data, int samples) { return jfauderr_unrecognised; }
static jfauderr readallpcmdata(jfaud_file *file, jfaud_buffer *data) { return jfauderr_unrecognised; }
static jfauderr rewindpcmdata(jfaud_file *file) { return jfauderr_unrecognised; }
static jfauderr getplaytime(jfaud_file *file, double *len) { return jfauderr_unrecognised; }
#else

#ifdef __WATCOMC__
#define __int32 int
#define __int16 short
#endif

#include "vorbis/codec.h"
#include "vorbis/vorbisfile.h"

static struct jfaud_init_params *params;

#if (DYNAMIC_VORBIS == 1)

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define DLLNAME "vorbisfile.dll"
#else
#include <dlfcn.h>
#define DLLNAME "libvorbisfile.so"
#endif

static struct {
#ifdef _WIN32
	HMODULE dll;
#else
	void *dll;
#endif
	int (*ov_open_callbacks)(void *datasource, OggVorbis_File *vf, char *initial, long ibytes, ov_callbacks callbacks);
	int (*ov_test_callbacks)(void *datasource, OggVorbis_File *vf, char *initial, long ibytes, ov_callbacks callbacks);
	int (*ov_test_open)(OggVorbis_File *vf);
	int (*ov_clear)(OggVorbis_File *vf);
	vorbis_info *(*ov_info)(OggVorbis_File *vf,int link);
	long (*ov_read)(OggVorbis_File *vf,char *buffer,int length, int bigendianp,int word,int sgned,int *bitstream);
	int (*ov_pcm_seek_lap)(OggVorbis_File *vf,ogg_int64_t pos);
	ogg_int64_t (*ov_pcm_total)(OggVorbis_File *vf,int i);
	double (*ov_time_total)(OggVorbis_File *vf,int i);
} vorbisfile = {NULL, NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

#define ov_open_callbacks vorbisfile.ov_open_callbacks
#define ov_test_callbacks vorbisfile.ov_test_callbacks
#define ov_test_open vorbisfile.ov_test_open
#define ov_clear vorbisfile.ov_clear
#define ov_info vorbisfile.ov_info
#define ov_read vorbisfile.ov_read
#define ov_pcm_seek_lap vorbisfile.ov_pcm_seek_lap
#define ov_pcm_total vorbisfile.ov_pcm_total
#define ov_time_total vorbisfile.ov_time_total

#define testifloaded() { if (vorbisfile.dll == NULL) return jfauderr_notinited; }
#else
#define testifloaded() { }
#endif


static size_t vfread(void *ptr, size_t size, size_t nmemb, void *datasource)
{ return params->call_read(((jfaud_file*)datasource)->fh, ptr, size*nmemb); }

static int vfseek(void *datasource, ogg_int64_t offset, int whence)
{ return params->call_seek(((jfaud_file*)datasource)->fh, (int)offset, whence); }

static int vfclose(void *datasource)
{ return params->call_close(((jfaud_file*)datasource)->fh); }

static long vftell(void *datasource)
{ return params->call_tell(((jfaud_file*)datasource)->fh); }

static ov_callbacks vfcallbacks = { vfread, vfseek, vfclose, vftell };



static jfauderr init(void)
{
	int er=0;
	void *sym;

	params = jfaud_getparams();

#if (DYNAMIC_VORBIS == 1)
#ifdef _WIN32
	vorbisfile.dll = LoadLibrary(DLLNAME);

#define gpa(d,x) sym = GetProcAddress(d,x); { \
		if (!sym) { \
			params->call_logmsg(jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
			er = 1; \
		} \
	}
#else
	vorbisfile.dll = dlopen(DLLNAME, RTLD_GLOBAL|RTLD_NOW);
	dlerror();

#define gpa(d,x) dlsym(d,x); { \
		const char *s; \
		if (s = dlerror()) { \
			params->call_logmsg(jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
			er = 1; \
		} \
	}
#endif
	if (!vorbisfile.dll) return jfauderr_notinited;
	
	ov_open_callbacks = (void*)gpa(vorbisfile.dll, "ov_open_callbacks");
	ov_test_callbacks = (void*)gpa(vorbisfile.dll, "ov_test_callbacks");
	ov_test_open = (void*)gpa(vorbisfile.dll, "ov_test_open");
	ov_clear = (void*)gpa(vorbisfile.dll, "ov_clear");
	ov_info = (void*)gpa(vorbisfile.dll, "ov_info");
	ov_read = (void*)gpa(vorbisfile.dll, "ov_read");
	ov_pcm_seek_lap = (void*)gpa(vorbisfile.dll, "ov_pcm_seek_lap");
	ov_pcm_total = (void*)gpa(vorbisfile.dll, "ov_pcm_total");
	ov_time_total = (void*)gpa(vorbisfile.dll, "ov_time_total");

	if (er) {
#ifdef _WIN32
		FreeLibrary(vorbisfile.dll);
#else
		dlclose(vorbisfile.dll);
#endif
		vorbisfile.dll = NULL;
		params->call_logmsg("Failed fetching required symbols from " DLLNAME ".");
		return jfauderr_notinited;
	}
#endif	

	return jfauderr_ok;
}

static jfauderr uninit(void)
{
#if (DYNAMIC_VORBIS == 1)
	if (vorbisfile.dll) {
#ifdef _WIN32
		FreeLibrary(vorbisfile.dll);
#else
		dlclose(vorbisfile.dll);
#endif
		vorbisfile.dll = NULL;
	}
#endif
	return jfauderr_ok;
}

static jfauderr testfile(jfaud_file *file, const char *sub)
{
	OggVorbis_File vf, *vorbf;
	int ret;

	testifloaded();

	params->call_seek(file->fh, 0, SEEK_SET);

	memset(&vf,0,sizeof(OggVorbis_File));

	/*
	ret = ov_test_callbacks((void*)file, &vf, NULL, 0, vfcallbacks);
	if (ret == OV_ENOTVORBIS) return jfauderr_unrecognised;
	else if (ret != 0) return jfauderr_genericerror;
	
	ret = ov_test_open(&vf);
	if (ret != 0) return jfauderr_genericerror;
	*/

	ret = ov_open_callbacks((void*)file, &vf, NULL, 0, vfcallbacks);
	if (ret == OV_ENOTVORBIS) return jfauderr_unrecognised;
	else if (ret != 0) return jfauderr_genericerror;

	vorbf = (OggVorbis_File *)malloc(sizeof(OggVorbis_File));
	memcpy(vorbf, &vf, sizeof(vf));

	file->type = jfaudfile_vorbis;
	file->readerinfo = (void*)vorbf;
	
	return jfauderr_ok;
}

static jfauderr cleanup(jfaud_file *file, char *noclose)
{
	testifloaded();
	
	if (file->readerinfo) {
		ov_clear((OggVorbis_File*)file->readerinfo);
		free(file->readerinfo);
	}
	file->readerinfo = NULL;
	*noclose = 1;

	return jfauderr_ok;
}

static jfauderr readpcmdata(jfaud_file *file, jfaud_buffer *data, int samples)
{
	OggVorbis_File *vorbf;
	vorbis_info *vi;
	unsigned int bytes, br;
	char *bp;
	int lstream=-1,stream;
	jfauderr err;
	
	testifloaded();

	if (file->type != jfaudfile_vorbis) return jfauderr_unrecognised;

	vorbf = (OggVorbis_File *)file->readerinfo;

	vi = ov_info(vorbf, -1);
	if (!vi) return jfauderr_genericerror;

	if ((err = jfaudbuffer_alloc(data, vi->rate, vi->channels, 2*8, samples)) != jfauderr_ok)
		return err;

	bp = (char *)data->data;
	bytes = vi->channels * 2 * samples;
	
	do {
		br = ov_read(vorbf, bp, bytes, 0,2,1,&stream);
		if (lstream != stream && lstream != -1)
			params->call_logmsg("Vorbis decoder warning: chained bitstream detected");
		lstream = stream;
		bytes -= br;
		bp += br;
	} while ((br > 0 || br == OV_HOLE) && bytes > 0);
	
	if (br < 0 && br != OV_HOLE) {
		// damaged file
		jfaudbuffer_free(data);
		return jfauderr_genericerror;
	}

	data->numsamples = ((long)bp - (long)data->data) / (vi->channels*2);
	
	return jfauderr_ok;
}

static jfauderr readallpcmdata(jfaud_file *file, jfaud_buffer *data)
{
	OggVorbis_File *vorbf;
	vorbis_info *vi;
	unsigned int bytes, br;
	char *bp;
	int lstream=-1,stream;
	jfauderr err;

	testifloaded();

	if (file->type != jfaudfile_vorbis) return jfauderr_unrecognised;

	vorbf = (OggVorbis_File *)file->readerinfo;

	vi = ov_info(vorbf, -1);
	if (!vi) return jfauderr_genericerror;

	if ((err = jfaudbuffer_alloc(data, vi->rate, vi->channels, 2*8, ov_pcm_total(vorbf, -1))) != jfauderr_ok)
		return err;
	
	bp = (char *)data->data;
	bytes = vi->channels * 2 * ov_pcm_total(vorbf, -1);

	do {
		br = ov_read(vorbf, bp, bytes, 0,2,1,&stream);
		if (lstream != stream && lstream != -1)
			params->call_logmsg("Vorbis decoder warning: chained bitstream detected while reading entire file");
		lstream = stream;
		bytes -= br;
		bp += br;
	} while ((br > 0 || br == OV_HOLE) && bytes > 0);

	if (br < 0 && br != OV_HOLE) {
		// damaged file
		jfaudbuffer_free(data);
		return jfauderr_genericerror;
	}
	
	return jfauderr_ok;
}

static jfauderr rewindpcmdata(jfaud_file *file)
{
	OggVorbis_File *vorbf;

	testifloaded();

	if (file->type != jfaudfile_vorbis) return jfauderr_unrecognised;

	vorbf = (OggVorbis_File *)file->readerinfo;

	if (ov_pcm_seek_lap(vorbf, 0)) return jfauderr_genericerror;

	return jfauderr_ok;
}

static jfauderr getpcmlength(jfaud_file *file, unsigned int *samples)
{
	OggVorbis_File *vorbf;
	double l;

	testifloaded();

	if (file->type != jfaudfile_vorbis) return jfauderr_unrecognised;

	vorbf = (OggVorbis_File *)file->readerinfo;

	l = ov_pcm_total(vorbf, -1);
	if (l == OV_EINVAL) return jfauderr_genericerror;
	
	*samples = l;
	return jfauderr_ok;
}

static jfauderr getplaytime(jfaud_file *file, double *len)
{
	OggVorbis_File *vorbf;
	double l;

	testifloaded();

	if (file->type != jfaudfile_vorbis) return jfauderr_unrecognised;

	vorbf = (OggVorbis_File *)file->readerinfo;

	l = ov_time_total(vorbf, -1);
	if (l == OV_EINVAL) return jfauderr_genericerror;
	
	*len = l;
	return jfauderr_ok;
}

#endif

jfaud_filereader reader_vorbis = {
	"OggVorbis",
	jfaudfile_vorbis,
	jfaudfiledata_pcm,

	init,
	uninit,
	testfile,
	cleanup,

	readpcmdata,
	readallpcmdata,
	rewindpcmdata,
	getpcmlength,
	getplaytime
};
