#include "internal.h"

#include <stdlib.h>
#include <stdio.h>

static struct jfaud_init_params *params;

#ifdef __GNUC__
#define PACKED __attribute__ ((packed))
#else
#define PACKED
#endif

#ifdef _MSC_VER
#pragma pack(1)
#endif
struct PACKED riffhead {
	char riff[4];
	unsigned long length;
	char type[4];
};

struct PACKED riffchunkhead {
	char id[4];
	unsigned long length;
};

struct PACKED wavefmt {
	short formattag;
	short channels;
	long samplespersec;
	long avgbytespersec;
	short blockalign;
};
#ifdef _MSC_VER
#pragma pack()
#endif

enum {
	wavefmt_pcm = 1
};

struct PACKED wavefmtpcm {
	short bitspersample;
};


struct wavinfo {
	unsigned long audiostart;	// offset to position of first audio sample
	unsigned long audiolength;	// byte count of audio data length

	unsigned long pcmposn;		// pcm playback position

	struct wavefmt wf;
	struct wavefmtpcm wfpcm;
};


static jfauderr init(void)
{
	params = jfaud_getparams();

	return jfauderr_ok;
}

static jfauderr uninit(void)
{
	return jfauderr_ok;
}

static jfauderr testfile(jfaud_file *file, const char *sub)
{
	struct riffhead rh;
	struct riffchunkhead rch;
	struct wavefmt wf;
	struct wavefmtpcm wfpcm;
	struct wavinfo *wav;
	char havefmt = 0;

	// FIXME: do more legality checking

	params->call_seek(file->fh, 0, SEEK_SET);

	if (params->call_read(file->fh, &rh, sizeof(rh)) != sizeof(rh))
		return jfauderr_genericerror;
	if (rh.riff[0] != 'R' || rh.riff[1] != 'I' || rh.riff[2] != 'F' || rh.riff[3] != 'F')
		return jfauderr_unrecognised;
	if (rh.type[0] != 'W' || rh.type[1] != 'A' || rh.type[2] != 'V' || rh.type[3] != 'E')
		return jfauderr_unrecognised;

	while (1) {
		if (params->call_read(file->fh, &rch, sizeof(rch)) != sizeof(rch))
			return jfauderr_genericerror;
		if (rch.id[0] == 'd' && rch.id[1] == 'a' && rch.id[2] == 't' && rch.id[3] == 'a') {
			if (!havefmt) {
				// missing the preceeding fmt chunk
				return jfauderr_genericerror;
			}

			break;	// now at the start of the waveform data
		} else if (rch.id[0] == 'f' && rch.id[1] == 'm' && rch.id[2] == 't' && rch.id[3] == ' ') {
			havefmt = 1;
			if (params->call_read(file->fh, &wf, sizeof(wf)) != sizeof(wf))
				return jfauderr_genericerror;

			if (wf.formattag == wavefmt_pcm) {
				if (params->call_read(file->fh, &wfpcm, sizeof(wfpcm)) != sizeof(wfpcm))
					return jfauderr_genericerror;
			} else {
				// unrecognised data format
				return jfauderr_genericerror;
			}
		} else {
			// unrecognised chunk. skip over it
			params->call_seek(file->fh, rch.length, SEEK_CUR);
		}
	}

	wav = (struct wavinfo *)malloc(sizeof(struct wavinfo));
	wav->audiostart = params->call_tell(file->fh);
	wav->audiolength = rch.length;
	wav->pcmposn = 0;
	wav->wf = wf;
	wav->wfpcm = wfpcm;

	file->type = jfaudfile_wave;
	file->readerinfo = (void*)wav;
	
	return jfauderr_ok;
}

static jfauderr cleanup(jfaud_file *file, char *noclose)
{
	if (file->readerinfo) free(file->readerinfo);
	file->readerinfo = NULL;
	*noclose = 0;

	return jfauderr_ok;
}

static jfauderr readpcmdata(jfaud_file *file, jfaud_buffer *data, int samples)
{
	struct wavinfo *wav;
	unsigned int samps, readbytes;
	jfauderr err;

	if (file->type != jfaudfile_wave) return jfauderr_unrecognised;

	wav = (struct wavinfo *)file->readerinfo;

	samps = (wav->audiolength - wav->pcmposn) / wav->wf.blockalign;
	if (samps == 0) {
		data->numsamples = 0;
		data->data = NULL;
		return jfauderr_ok;
	}
	
	if (samps < samples) samples = samps;

	if ((err = jfaudbuffer_alloc(data, wav->wf.samplespersec, wav->wf.channels, wav->wfpcm.bitspersample, samples)) != jfauderr_ok)
		return err;

	// FIXME: check read length
	readbytes = params->call_read(file->fh, data->data, samples * wav->wf.blockalign);
	if (readbytes < 0) return jfauderr_genericerror;
	else if (readbytes == 0) {	// premature eof
		jfaudbuffer_free(data);
		return jfauderr_ok;
	}

	wav->pcmposn += readbytes;

	return jfauderr_ok;
}

static jfauderr readallpcmdata(jfaud_file *file, jfaud_buffer *data)
{
	struct wavinfo *wav;
	unsigned int samps, readbytes;
	jfauderr err;

	if (file->type != jfaudfile_wave) return jfauderr_unrecognised;

	wav = (struct wavinfo *)file->readerinfo;

	samps = (wav->audiolength - wav->pcmposn) / wav->wf.blockalign;
	if (samps == 0) {
		data->numsamples = 0;
		data->data = NULL;
		return jfauderr_ok;
	}
	
	if ((err = jfaudbuffer_alloc(data, wav->wf.samplespersec, wav->wf.channels, wav->wfpcm.bitspersample, samps)) != jfauderr_ok)
		return err;

	// FIXME: check read length
	readbytes = params->call_read(file->fh, data->data, samps * wav->wf.blockalign);
	if (readbytes < 0) return jfauderr_genericerror;
	else if (readbytes == 0) {	// premature eof
		jfaudbuffer_free(data);
		return jfauderr_ok;
	}

	wav->pcmposn += readbytes;

	return jfauderr_ok;
}

static jfauderr rewindpcmdata(jfaud_file *file)
{
	struct wavinfo *wav;
	
	if (file->type != jfaudfile_wave) return jfauderr_unrecognised;

	wav = (struct wavinfo *)file->readerinfo;

	wav->pcmposn = 0;
	params->call_seek(file->fh, wav->audiostart, SEEK_SET);

	return jfauderr_ok;
}

static jfauderr getpcmlength(jfaud_file *file, unsigned int *samples)
{
	struct wavinfo *wav;
	
	if (file->type != jfaudfile_wave) return jfauderr_unrecognised;

	wav = (struct wavinfo *)file->readerinfo;

	*samples = wav->audiolength / wav->wf.blockalign;

	return jfauderr_ok;
}

static jfauderr getplaytime(jfaud_file *file, double *len)
{
	struct wavinfo *wav;
	
	if (file->type != jfaudfile_wave) return jfauderr_unrecognised;

	wav = (struct wavinfo *)file->readerinfo;

	*len = (double)(wav->audiolength / wav->wf.blockalign) / (double)wav->wf.samplespersec;

	return jfauderr_ok;
}


jfaud_filereader reader_wave = {
	"RIFF Wave",
	jfaudfile_wave,
	jfaudfiledata_pcm,

	init,
	uninit,
	testfile,
	cleanup,

	readpcmdata,
	readallpcmdata,
	rewindpcmdata,
	getpcmlength,
	getplaytime
};
