#ifndef __jfaud_h__
#define __jfaud_h__

#include "jfauderrs.h"


enum {
	JFAudFH_startoffs = 1,
	JFAudFH_length = 2
};

typedef struct {
	int fh;
	void *user;	// used by the callback to remember information about the file if it needs
} JFAudFH;

typedef struct {
	unsigned int samplerate;
	unsigned char channels;
	unsigned char bitspersample;
	unsigned char bigendian;
} JFAudRawFormat;

typedef struct {
	char *wavedev;			// OpenAL audio device string
		// NULL = no waveform audio
		// empty string = default
		// device name
	char *mididev;			// MIDI device string
		// NULL = no MIDI
		// empty string = midi mapper
		// (win32) device name
		// "*adlibemu" = Ken's adlib emulator (requires a pcm device)
	char *cdadev;			// CD-A device string
		// NULL = no CD-A
		// empty string = first CDROM drive
		// (win32) drive letter
		// (*nix) cdrom device

	unsigned int maxvoices;			// maximum number of sounds to play simultaneously
	unsigned int samplerate;		// sampling frequency to try and use
	unsigned int synchronous;		// false = spawn update thread
	
	int (*cb_open)(const char *fn, const char *subfn, JFAudFH *h, JFAudRawFormat **raw);	// file opening callback
		// A word about the 'raw' parameter:
		//   If the file is opened and JFAud should just treat the samples it reads as headerless
		//   PCM, '*raw' should be set to a raw format descriptor, usually allocated in 'h->user',
		//   otherwise it should be set to NULL
	int (*cb_close)(JFAudFH *h);				// file closing callback
	int (*cb_read)(JFAudFH *h, void *buf, int len);	// file reading callback
	int (*cb_seek)(JFAudFH *h, int off, int whence);	// file seeking callback
	int (*cb_tell)(JFAudFH *h);				// file telling callback
	int (*cb_filelen)(JFAudFH *h);				// file length callback

	void (*cb_logmsg)(const char *str);	// log message output callback
} JFAudCfg;

enum {
	JFAudProp_null = 0,
	JFAudProp_gain,		// float	(listener, channel, music)
	JFAudProp_pitch,	// float	(channel)
	JFAudProp_loop,		// int(bool)	(channel, music)
	JFAudProp_rolloff,	// float	(channel)	1.0 = normal, 0.0 = 2d sound
	JFAudProp_position,	// vector	(listener, channel)
	JFAudProp_velocity,	// vector	(listener, channel)
	JFAudProp_orient,	// vector2	(listener)
	JFAudProp_posrel,	// int(bool)	(channel)	0=listener-relative, !0=world-relative
	JFAudProp_mindist,	// float	(channel)
	JFAudProp_fx		// int		(channel)	1=nearest-resample (init-only option)
				//				2=stereo-to-mono (init-only option)
};
		
typedef struct {
	int prop;
	union {
		int i;
		float f;
		float v[3];
		float v2[6];
	} val;
} JFAudProp;

typedef enum {
	JFAudPlayMode_stopped,
	JFAudPlayMode_paused,
	JFAudPlayMode_playing
} JFAudPlayMode;

typedef enum {
	JFAudMetaInfo_title,
	JFAudMetaInfo_artist
} JFAudMetaInfo;

typedef enum {
	JFAudInit_base,
	JFAudInit_wave,
	JFAudInit_midi,
	JFAudInit_cda
} JFAudInit;

jfauderr JFAud_init(JFAudCfg *cfg);
jfauderr JFAud_uninit(void);
jfauderr JFAud_isinited(JFAudInit what);

#define DEBUGLVL_GENERAL 1
#define DEBUGLVL_WORDY   64
#define DEBUGLVL_TRACE   128
#define DEBUGLVL_ALL     256
jfauderr JFAud_setdebuglevel(int l);

jfauderr JFAud_update(void);	// call to run an update cycle in synchronous mode

#define JFAudDefaultStreamThresh (4.f)
jfauderr JFAud_setstreamthresh(float secs);	// number of seconds before streaming should be used

#define JFAudDefaultStreamSamples (10240)
jfauderr JFAud_setstreamsamples(unsigned int samples);

// JFAud_playsound
//   *handle - pointer to integer that will hold the sound (channel) handle
//   *fn,*subfn - filename of the sound
//   initial - initial playback mode (ie. paused or playing)
//   nprops,props - properties for initialisation before sound is loaded
jfauderr JFAud_playsound(int *handle, const char *fn, const char *subfn, int priority, JFAudPlayMode initial,
		int nprops, JFAudProp *props);

// JFAud_playrawsound
//   NOTE: sounds played this way are not handled by the buffer managment code yet
//   *handle - pointer to integer that will hold the sound (channel) handle
//   data,dataleng,samplerate,channels,bitspersample - vital stats of the sound
//   initial - initial playback mode (ie. paused or playing)
jfauderr JFAud_playrawsound(int *handle, const void *data, unsigned int dataleng,
		JFAudRawFormat *raw, int priority, JFAudPlayMode initial);

// JFAud_playmusic
//   *fn,*subfn - filename of the music
//   initial - initial playback mode
jfauderr JFAud_playmusic(const char *fn, const char *subfn, JFAudPlayMode initial, int nprops, JFAudProp *props);

jfauderr JFAud_killsound(int handle);

jfauderr JFAud_setlistenerprops(int nprops, JFAudProp *props);
jfauderr JFAud_setsoundprops(int handle, int nprops, JFAudProp *props);
jfauderr JFAud_setmusicprops(int nprops, JFAudProp *props);

jfauderr JFAud_getsoundplaymode(int handle, JFAudPlayMode *mode);
jfauderr JFAud_setsoundplaymode(int handle, JFAudPlayMode mode);

jfauderr JFAud_getmusicplaymode(JFAudPlayMode *mode);
jfauderr JFAud_setmusicplaymode(JFAudPlayMode mode);


#endif	// __jfaud_h__

