#include <stdlib.h>
#include <string.h>

#include "internal.h"

static char inited = 0;

typedef struct {
	unsigned long age;
	File *data;			// NULL for non-streamed channels
	fileident ident;		// if (ident.a.flags&1) is set, this is a doublespeed-hacked channel
	int priority;

	char stream,loop,fx,filler;
	
	int outhandle;			// handle issued by the output device
} WaveChannel;

typedef struct {
	MidiSequencer seq;
	fileident ident;
	int outhandle;

	unsigned delta, len;
	unsigned char const *data;
	unsigned char cmd, paused, endofsong;
} MidiChannel;

typedef struct {
	int starttrack,endtrack;
	char inuse,loop;
} CdaChannel;

static unsigned int streamsamples = JFAudDefaultStreamSamples;
static WaveChannel *wavechans = NULL;
static MidiChannel midichan;
static CdaChannel  cdachan;

jfauderr JFAud_setstreamsamples(unsigned int samples)
{
	streamsamples = samples;
	return jfauderr_noerror;
}


//{{{ channel_init
jfauderr channel_init(void)
{
	int i;
	
	if (inited) return jfauderr_alreadyinited;

	if (conf.wavedev) {
		wavechans = (WaveChannel *)calloc(conf.maxvoices, sizeof(WaveChannel));
		if (!wavechans) return jfauderr_outofmemory;

		for (i=conf.maxvoices-1; i>=0; i--) wavechans[i].outhandle = -1;
	}

	midichan.outhandle = -1;
	cdachan.inuse = 0;

	inited = 1;

	return jfauderr_noerror;
}
//}}}

//{{{ channel_uninit
jfauderr channel_uninit(void)
{
	int i;
	
	if (!inited) return jfauderr_notinited;

	if (wavechans) {
		for (i=conf.maxvoices-1; i>=0; i--)
			if (wavechans[i].outhandle >= 0)
				channel_release( CHANTYPE_WAVE | i );
		free(wavechans);
		wavechans = NULL;
	}

	if (midichan.outhandle >= 0)
		channel_release( CHANTYPE_MIDI );

	if (cdachan.inuse)
		channel_release( CHANTYPE_CDA );

	inited = 0;

	return jfauderr_noerror;
}
//}}}

//{{{ channel_updateonewave
static jfauderr channel_updateonewave(WaveChannel *ch)
{
	jfauderr err;
	PcmBuffer buf;

	err = openal_needsdata(ch->outhandle);
	if (err != jfauderr_ok) return err;

	err = filereaders_wave_readsamples(ch->data, streamsamples, &buf, ch->loop);
	if (err != jfauderr_noerror) return err;
	if (buf.numsamples == 0) return jfauderr_noerror;

	if (ch->data->ident.a.flags & fileidentflag_fxstertomono)
		if ((err = fx_stertomono(&buf)) != jfauderr_noerror) {
			pcmbuffer_free(&buf);
			return err;
		}
	if (ch->data->ident.a.flags & fileidentflag_fxnearest)
		if ((err = fx_nearest(&buf)) != jfauderr_noerror) {
			pcmbuffer_free(&buf);
			return err;
		}
	if (ch->data->ident.a.flags & fileidentflag_doublespeed)
		buf.samplerate <<= 1;

	err = openal_preparebuffer(&buf);
	if (err != jfauderr_noerror) return err;
	
	err = openal_queuedata(ch->outhandle, &buf);
	return err;
}
//}}}

//{{{ channel_updateonemidi
static jfauderr channel_updateonemidi(MidiChannel *ch)
{
	jfauderr err;
	unsigned int tickamt;

	if (ch->endofsong) return jfauderr_noerror;

	err = midiout_needsdata(ch->outhandle);
	if (err != jfauderr_noerror) return err;

	if (ch->cmd == 0) {
		switch (midiseq_getevent(&ch->seq, &ch->delta, &ch->cmd, &ch->len, &ch->data)) {
			case 1: ch->endofsong = 1; return jfauderr_noerror;	// end of song
			case 0: break;
			default: ch->cmd = 0; return jfauderr_genericerror;
		}
	}

	tickamt = ch->seq.midifile->ppqn;

	if (ch->paused || ch->delta > tickamt) {
		if (!ch->paused) ch->delta -= tickamt;
		err = midiout_bufferevent(ch->outhandle, tickamt, 0,0,NULL);
		if (err != jfauderr_noerror) return err;

		return midiout_finish(ch->outhandle);
	}

	do {
		err = midiout_bufferevent(ch->outhandle, ch->delta, ch->cmd, ch->len, ch->data);
		if (err == jfauderr_genericerror) break;	// buffer full

		tickamt -= ch->delta;

		switch (midiseq_getevent(&ch->seq, &ch->delta, &ch->cmd, &ch->len, &ch->data)) {
			case 1: ch->endofsong = 1; ch->cmd = 0;
			case 0: break;
			default: ch->cmd = 0; return jfauderr_noerror;
		}
	} while (ch->delta <= tickamt);

	return midiout_finish(ch->outhandle);
}
//}}}

//{{{ channel_updateonecda
static jfauderr channel_updateonecda(CdaChannel *ch)
{
	// FIXME: do loop handling!
	return jfauderr_noerror;
}
//}}}

//{{{ channel_update
jfauderr channel_update(void)
{
	jfauderr err;
	int i;
	
	if (wavechans) {
		err = openal_update();
		if (err != jfauderr_noerror) return err;

		for (i=conf.maxvoices-1; i>=0; i--) {
			if (wavechans[i].outhandle < 0) continue;
			wavechans[i].age++;
			if (!wavechans[i].data) continue;

			err = channel_updateonewave(&wavechans[i]);
			if (err != jfauderr_ok && err != jfauderr_genericerror) return err;
		}
	}

	if (midichan.outhandle >= 0) {
		err = channel_updateonemidi(&midichan);
		if (err != jfauderr_ok && err != jfauderr_genericerror) return err;
	}

	if (cdachan.inuse) {
		err = channel_updateonecda(&cdachan);
		if (err != jfauderr_ok && err != jfauderr_genericerror) return err;
	}

	return jfauderr_noerror;
}
//}}}

//{{{ findwavechan
static int findwavechan(int prio)
{
	int lowestchan, lowestprio, i;
	unsigned long oldage;

	if (!wavechans) return -1;

	lowestchan = conf.maxvoices;
	lowestprio = 0x7fffffffl;
	oldage = 0;

	for (i=conf.maxvoices-1; i>=0; i--) {
		if (wavechans[i].outhandle < 0) {
			lowestchan = i;
			lowestprio = 0;
			break;
		}
		if (wavechans[i].priority != ReservedPriority &&		// never turf out reserved channels
		    (wavechans[i].priority < lowestprio ||			// lower prio channels are first to go
		     (wavechans[i].priority == lowestprio && wavechans[i].age>oldage)	// oldest chan of same prio goes first
		    )
		   ) {
			debuglog(DEBUGLVL_TRACE,jsprintf("found kickout candidate %d: prio %d->%d age %ld->%ld",
							i,lowestprio,wavechans[i].priority,oldage,wavechans[i].age));
			lowestchan = i;
			lowestprio = wavechans[i].priority;
			oldage     = wavechans[i].age;
		}
	}

	if (prio < lowestprio && prio != ReservedPriority) return -1;
	return lowestchan;
}
//}}}

//{{{ channel_allocwave
jfauderr channel_allocwave(jfaudchan *chan, fileident ident, int priority)
{
	int ch;
	jfauderr err;
	Buffer *buf = NULL;
	
	if (!inited) return jfauderr_notinited;

	err = buffer_fetch(ident, &buf);
	if (err != jfauderr_ok) return err;
	if (!buf) return jfauderr_notloaded;
	if (buf->type != buffertype_pcm) {
		buffer_release(ident);
		return jfauderr_notcompatible;
	}

	ch = findwavechan(priority);
	if (ch < 0) {
		buffer_release(ident);
		return jfauderr_nofreechans;
	}

	if (wavechans[ch].outhandle >= 0) {
		debuglog(DEBUGLVL_WORDY,jsprintf("Kicking out wave channel %d (prio=%d, age=%ld)",
					ch,wavechans[ch].priority,wavechans[ch].age));
		channel_release(CHANTYPE_WAVE | ch);
	}

	err = openal_allocsource(&wavechans[ch].outhandle,0);
	if (err != jfauderr_ok) {
		buffer_release(ident);
		return err;
	}

	wavechans[ch].age = 0;
	wavechans[ch].data = NULL;
	wavechans[ch].ident = ident;
	wavechans[ch].stream = 0;
	wavechans[ch].priority = priority;

	err = openal_queuedata(wavechans[ch].outhandle, &buf->buf.pcm);
	if (err != jfauderr_ok) {
		channel_release(CHANTYPE_WAVE | ch);
		// channel_release has released the buffer for us
		return err;
	}

	*chan = (jfaudchan)(CHANTYPE_WAVE | ch);
	return jfauderr_noerror;
}
//}}}

//{{{ channel_allocwavestream
jfauderr channel_allocwavestream(jfaudchan *chan, File *file, int priority)
{
	int ch;
	jfauderr err;
	
	if (!inited) return jfauderr_notinited;

	ch = findwavechan(priority);
	if (ch < 0) return jfauderr_nofreechans;

	if (wavechans[ch].outhandle >= 0) {
		debuglog(DEBUGLVL_WORDY,jsprintf("Kicking out wave channel %d (prio=%d, age=%ld)",
					ch,wavechans[ch].priority,wavechans[ch].age));
		channel_release(CHANTYPE_WAVE | ch);
	}

	err = openal_allocsource(&wavechans[ch].outhandle,1);
	if (err != jfauderr_ok) return err;

	wavechans[ch].age = 0;
	wavechans[ch].data = file;
	wavechans[ch].ident = file->ident;
	wavechans[ch].stream = 1;
	wavechans[ch].priority = priority;
/*
	err = channel_updateone(&wavechans[ch]);
	if (err != jfauderr_ok) {
		channel_release(CHANTYPE_WAVE | ch);
		return err;
	}
*/
	*chan = (jfaudchan)(CHANTYPE_WAVE | ch);
	return jfauderr_noerror;
}
//}}}

//{{{ channel_allocmidi
jfauderr channel_allocmidi(jfaudchan *chan, fileident ident)
{
	jfauderr err;
	Buffer *buf = NULL;
	mididevtype devtype;

	if (!inited) return jfauderr_notinited;

	err = buffer_fetch(ident, &buf);
	if (err != jfauderr_ok) return err;
	if (!buf) return jfauderr_notloaded;
	if (buf->type != buffertype_midi) {
		buffer_release(ident);
		return jfauderr_notcompatible;
	}

	if (midichan.outhandle >= 0) {
		buffer_release(ident);
		return jfauderr_nofreechans;
	}

	err = midiseq_new(&midichan.seq, &buf->buf.midi);
	if (err != jfauderr_ok) {
		buffer_release(ident);
		return err;
	}
	
	err = midiout_allocdevice(&midichan.outhandle);
	if (err != jfauderr_ok) {
		midiseq_free(&midichan.seq);
		buffer_release(ident);
		return err;
	}

	midichan.ident = ident;
	midichan.cmd = 0;
	midichan.paused = 1;
	midichan.endofsong = 0;

	err = midiout_settimebase(midichan.outhandle, buf->buf.midi.ppqn);
	if (err != jfauderr_noerror) {
		channel_release(CHANTYPE_MIDI);
		return err;
	}
	
	err = midiout_getdevicetype(midichan.outhandle, &devtype);
	if (err != jfauderr_noerror) {
		channel_release(CHANTYPE_MIDI);
		return err;
	}
	switch (devtype) {
		case mididevtype_fmopl2:
		case mididevtype_fmopl3:
			err = midiseq_setdevice(&midichan.seq, midiseqdev_adlib);
			break;
		case mididevtype_genmidi:
		default:
			err = midiseq_setdevice(&midichan.seq, midiseqdev_genmidi);
			break;
	}
	if (err != jfauderr_noerror) {
		channel_release(CHANTYPE_MIDI);
		return err;
	}

	*chan = (jfaudchan)(CHANTYPE_MIDI);
	return jfauderr_noerror;
}
//}}}

//{{{ channel_alloccda
jfauderr channel_alloccda(jfaudchan *chan, int starttrack, int endtrack)
{
	if (!inited) return jfauderr_notinited;

	if (cdachan.inuse) return jfauderr_nofreechans;

	if (starttrack >= 0 && starttrack >= cda_numtracks) return jfauderr_openerror;	// "file not found"
	if (endtrack >= 0   && endtrack >= cda_numtracks)   return jfauderr_openerror;	// "file not found"

	cdachan.starttrack = starttrack;
	cdachan.endtrack   = endtrack;
	cdachan.loop       = 0;
	cdachan.inuse      = 1;

	*chan = (jfaudchan)(CHANTYPE_CDA);
	return jfauderr_noerror;
}
//}}}

//{{{ channel_release
jfauderr channel_release(jfaudchan chan)
{
	unsigned int dachan;
	
	if (!inited) return jfauderr_notinited;

	dachan = chan & ~CHANTYPE_MASK;
	switch (chan & CHANTYPE_MASK) {
		case CHANTYPE_WAVE:
			if (dachan >= conf.maxvoices) break;
			if (wavechans[dachan].outhandle < 0) break;

			openal_releasesource(wavechans[dachan].outhandle);
			if (!wavechans[dachan].stream) buffer_release(wavechans[dachan].ident);
			if (wavechans[dachan].data) soundfile_close(wavechans[dachan].data); // FIXME: close file

			memset(&wavechans[dachan], 0, sizeof(WaveChannel));
			wavechans[dachan].outhandle = -1;

			return jfauderr_noerror;

		case CHANTYPE_MIDI:
			if (midichan.outhandle < 0) break;

			midiout_releasedevice(midichan.outhandle);
			if (midichan.seq.seqtrack) {
				if (midichan.seq.midifile) buffer_release(midichan.ident);
				midiseq_free(&midichan.seq);
			}
			midichan.outhandle = -1;

			return jfauderr_noerror;

		case CHANTYPE_CDA:
			if (!cdachan.inuse) break;

			cda_stop();
			cdachan.inuse = 0;

			return jfauderr_noerror;

		default: break;
	}

	return jfauderr_badhandle;
}
//}}}

//{{{ channel_setprops
jfauderr channel_setprops(jfaudchan chan, int nprops, JFAudProp *props)
{
	unsigned int dachan;
	int i;
	jfauderr err;

	// we can modify the props array passed to us since the user-callable function duplicates
	// it and filters out things we don't handle anyway.
	
	if (!inited) return jfauderr_notinited;

	dachan = chan & ~CHANTYPE_MASK;
	switch (chan & CHANTYPE_MASK) {
		case CHANTYPE_WAVE:
			if (dachan >= conf.maxvoices) break;
			if (wavechans[dachan].outhandle < 0) break;

			for (i=0;i<nprops;i++) {
				switch (props[i].prop) {
					case JFAudProp_loop: wavechans[dachan].loop = props[i].val.i; break;
					case JFAudProp_fx: wavechans[dachan].fx = props[i].val.i; break;
					case JFAudProp_pitch:
						if (wavechans[dachan].ident.a.flags & fileidentflag_doublespeed)
							props[i].val.f /= 2.0;

						if (props[i].val.f > 2.0)
							props[i].val.f = 2.0;	// we clamp, or we could baulk
							//return jfauderr_genericerror;
						break;
					default: break; 
				}
			}
			return openal_setsourceprops(wavechans[dachan].outhandle, nprops, props);

		case CHANTYPE_MIDI:
			if (midichan.outhandle < 0) break;

			for (i=0;i<nprops;i++) {
				switch (props[i].prop) {
					case JFAudProp_gain:
						// FIXME: implement this
						break;
					case JFAudProp_loop:
						err = midiseq_setloop(&midichan.seq, props[i].val.i);
						if (err != jfauderr_noerror) {
							debuglog(DEBUGLVL_GENERAL,"failed setting loop property on midiseq");
							return err;
						}
						break;
					default: break;
				}
			}

			return jfauderr_noerror;

		case CHANTYPE_CDA:
			if (!cdachan.inuse) break;

			for (i=0;i<nprops;i++) {
				switch (props[i].prop) {
					case JFAudProp_gain:
						// FIXME: implement this
						break;
					case JFAudProp_loop:
						cdachan.loop = props[i].val.i;
						break;
					default: break;
				}
			}

			return jfauderr_noerror;

		default: break;
	}

	return jfauderr_badhandle;
}
//}}}

//{{{ channel_getplaymode
jfauderr channel_getplaymode(jfaudchan chan, JFAudPlayMode *mode)
{
	unsigned int dachan;
	
	if (!inited) return jfauderr_notinited;

	dachan = chan & ~CHANTYPE_MASK;
	switch (chan & CHANTYPE_MASK) {
		case CHANTYPE_WAVE:
			if (dachan >= conf.maxvoices) break;
			if (wavechans[dachan].outhandle < 0) break;

			return openal_getsourceplaymode(wavechans[dachan].outhandle, mode);

		case CHANTYPE_MIDI:
			if (midichan.outhandle < 0) break;

			if (midichan.endofsong) *mode = JFAudPlayMode_stopped;
			else if (midichan.paused) *mode = JFAudPlayMode_paused;
			else *mode = JFAudPlayMode_playing;

			return jfauderr_noerror;

		case CHANTYPE_CDA: {
			jfauderr err;
			cdastatus st;

			if (!cdachan.inuse) break;

			err = cda_getstatus(&st);
			if (err != jfauderr_noerror) return err;
			switch (st) {
				case cdastatus_playing: *mode = JFAudPlayMode_playing;
				case cdastatus_paused:  *mode = JFAudPlayMode_paused;
				default: *mode = JFAudPlayMode_stopped;
			}

			return jfauderr_noerror;
		}

		default: break;
	}

	return jfauderr_badhandle;
}
//}}}

//{{{ channel_setplaymode
jfauderr channel_setplaymode(jfaudchan chan, JFAudPlayMode mode)
{
	unsigned int dachan;
	
	if (!inited) return jfauderr_notinited;

	dachan = chan & ~CHANTYPE_MASK;
	switch (chan & CHANTYPE_MASK) {
		case CHANTYPE_WAVE:
			if (dachan >= conf.maxvoices) break;
			if (wavechans[dachan].outhandle < 0) break;

			return openal_setsourceplaymode(wavechans[dachan].outhandle, mode);

		case CHANTYPE_MIDI:
			if (midichan.outhandle < 0) break;

			if (mode == JFAudPlayMode_stopped)
				midichan.endofsong = 1;
			else if (mode == JFAudPlayMode_paused)
				midichan.paused = 1;
			else
				midichan.paused = 0;

			return jfauderr_noerror;

		case CHANTYPE_CDA: {
			jfauderr err;
			cdastatus st;

			if (!cdachan.inuse) break;

			err = cda_getstatus(&st);
			if (err != jfauderr_noerror) return err;

			if (mode == JFAudPlayMode_stopped) cda_stop();
			else if (mode == JFAudPlayMode_paused) {
				if (st == cdastatus_playing)
					cda_pause(1);
			} else {
				if (st == cdastatus_paused)
					cda_pause(0);
				else {
					int startframe = -1, endframe = -1;
					if (cdachan.starttrack >= 0) startframe = cda_trackinfo[cdachan.starttrack].start;
					if (cdachan.endtrack   >= 0) endframe   = cda_trackinfo[cdachan.endtrack].start +
									cda_trackinfo[cdachan.endtrack].length - 1;
					if (startframe > endframe) {
						int i = startframe;
						startframe = endframe;
						endframe = i;
					}
					cda_play(startframe, endframe);
				}
			}	

			return jfauderr_noerror;
		}

		default: break;
	}

	return jfauderr_badhandle;
}
//}}}

// vim:fdm=marker:fdc=2:
