#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#ifndef _WIN32
#include <unistd.h>
#endif

#include "internal.h"
#include "thread.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif


static int inited = 0;
static jfauderr subsyserr[3] = {jfauderr_notinited,jfauderr_notinited,jfauderr_notinited};
static int debuglevel = 0;
static float streamthresh = JFAudDefaultStreamThresh;
JFAudCfg conf;

static int stdopen(const char *fn, const char *subfn UNUSED, JFAudFH *h, JFAudRawFormat **raw) {
	*raw = NULL;
	memset(h, 0, sizeof(JFAudFH));
	h->fh = open(fn, O_RDONLY|O_BINARY);
	return h->fh;
}
static int stdclose(JFAudFH *h) { return close(h->fh); }
static int stdread(JFAudFH *h, void *buf, int len) { return read(h->fh, buf, len); }
static int stdseek(JFAudFH *h, int off, int whence) { return lseek(h->fh, off, whence); }
static int stdtell(JFAudFH *h) { return lseek(h->fh,0,SEEK_CUR); }
static int stdfilelen(JFAudFH *h) {
#if defined _MSC_VER || defined __WATCOMC__
	return filelength(h->fh);
#else
	int len, cur = lseek(h->fh,0,SEEK_CUR);
	len = lseek(h->fh,0,SEEK_END);
	lseek(h->fh,cur,SEEK_SET);
	return len;
#endif
}
static void stdlogmsg(const char *str) { fputs(str, stderr); fputc('\n',stderr); }

static jfaudchan musicchan = -1;


jfauderr JFAud_init(JFAudCfg *cfg)
{
	jfauderr err;
	
	if (inited) return jfauderr_alreadyinited;

	memset(&conf, 0, sizeof(conf));
	
	conf.wavedev = (cfg->wavedev) ? strdup(cfg->wavedev) : NULL;
	conf.mididev = (cfg->mididev) ? strdup(cfg->mididev) : NULL;
	conf.cdadev  = (cfg->cdadev)  ? strdup(cfg->cdadev)  : NULL;

	conf.maxvoices = max(4,cfg->maxvoices);
	conf.samplerate = (cfg->samplerate <= 0) ? cfg->samplerate : 22050;
	conf.synchronous = (cfg->synchronous != 0);

	conf.cb_open  = (cfg->cb_open)  ? cfg->cb_open  : stdopen;
	conf.cb_close = (cfg->cb_close) ? cfg->cb_close : stdclose;
	conf.cb_read  = (cfg->cb_read)  ? cfg->cb_read  : stdread;
	conf.cb_seek  = (cfg->cb_seek)  ? cfg->cb_seek  : stdseek;
	conf.cb_tell  = (cfg->cb_tell)  ? cfg->cb_tell  : stdtell;
	conf.cb_filelen = (cfg->cb_filelen) ? cfg->cb_filelen : stdfilelen;

	conf.cb_logmsg = (cfg->cb_logmsg) ? cfg->cb_logmsg : stdlogmsg;

	inited = 1;
	
	err = filereaders_init();
	if (err != jfauderr_noerror) {
		JFAud_uninit();
		return err;
	}

	if (conf.wavedev) {
		subsyserr[0] = openal_init();
		if (subsyserr[0] != jfauderr_noerror) {
			free(conf.wavedev);
			conf.wavedev = NULL;
		}
	}
			
	if (conf.mididev) {
		subsyserr[1] = midiout_init();
		if (subsyserr[1] != jfauderr_noerror) {
			free(conf.mididev);
			conf.mididev = NULL;
		}
	}

	if (conf.cdadev) {
		subsyserr[2] = cda_opendevice();
		if (subsyserr[2] != jfauderr_noerror) {
			free(conf.cdadev);
			conf.cdadev = NULL;
		} else {
			cda_querydisc();
		}
	}

	err = buffer_init();
	if (err != jfauderr_noerror) {
		JFAud_uninit();
		return err;
	}

	err = channel_init();
	if (err != jfauderr_noerror) {
		JFAud_uninit();
		return err;
	}

	return jfauderr_noerror;
}

jfauderr JFAud_uninit(void)
{
	if (!inited) return jfauderr_notinited;

	channel_uninit();
	buffer_uninit();
	if (conf.cdadev)  cda_closedevice();
	if (conf.mididev) midiout_uninit();
	if (conf.wavedev) openal_uninit();
	filereaders_uninit();

	if (conf.wavedev) free(conf.wavedev);
	if (conf.mididev) free(conf.mididev);
	if (conf.cdadev)  free(conf.cdadev);

	memset(&conf, 0, sizeof(conf));

	inited = 0;

	return jfauderr_noerror;
}

jfauderr JFAud_isinited(JFAudInit what)
{
	if (what == JFAudInit_base) return inited ? jfauderr_noerror : jfauderr_notinited;
	if (what >= JFAudInit_wave && what <= JFAudInit_cda) return subsyserr[what - JFAudInit_wave];
	return jfauderr_genericerror;
}

jfauderr JFAud_setdebuglevel(int l)
{
	debuglevel = l;
	return jfauderr_noerror;
}

jfauderr JFAud_setstreamthresh(float secs)
{
	if (secs < 0.f) return jfauderr_genericerror;
	streamthresh = secs;
	return jfauderr_noerror;
}


static int shouldstream(float time)
{
	if (time < 0.0) return 1;
	if (time < streamthresh) return 0;
	if (time >= streamthresh && time < streamthresh * 1.5) return 0;
	return 1;
}

// mode == 0: regular sound, mode == 1: music
static jfauderr playsound(int *handle, const char *fn, const char *subfn, int priority, JFAudPlayMode initial,
		int nprops, JFAudProp *props, int mode)
{
	jfauderr err;
	File *fil;
	fileident id;
	jfaudchan chan;
	int i,stream=0;
	int soundtype = -1;

	int danprops = 0;
	JFAudProp *daprops;

	id = hashfn(fn,subfn);
	
	if (!props) nprops = 0;
	else {
		if (nprops < 0) nprops = 0;

		daprops = (JFAudProp *)calloc(nprops,sizeof(JFAudProp));
		if (!daprops) {
			debuglog(DEBUGLVL_TRACE,"calloc() failed");
			return jfauderr_outofmemory;
		}
		danprops = 0;

		for (i=0;i<nprops;i++) {
			daprops[danprops] = props[i];
			switch (props[i].prop) {
				case JFAudProp_rolloff:
				case JFAudProp_position:
				case JFAudProp_velocity:
				case JFAudProp_posrel:
				case JFAudProp_mindist:
				case JFAudProp_gain:
				case JFAudProp_loop:
					danprops++;
					break;
				case JFAudProp_fx:
					id.a.flags |= (daprops[danprops].val.i & 1) ? fileidentflag_fxnearest    : 0;
					id.a.flags |= (daprops[danprops].val.i & 2) ? fileidentflag_fxstertomono : 0;
					danprops++;
					break;
				case JFAudProp_pitch:
					if (props[i].val.f >= 2.0) {
						id.a.flags |= fileidentflag_doublespeed;
						//daprops[danprops].val.f = props[i].val.f / 2.0;
					}
					danprops++;
					break;
				default: break;
			}
		}
	}

	if (fn && !strcasecmp(fn,"*CDA")) soundtype = -2;	// to bypass any file access stuff for CD audio

	if (soundtype == -1 && buffer_isavailable(id, &soundtype) == jfauderr_notloaded) {
		Buffer buf;
		float playlen = 0.0;

		memset(&buf, 0, sizeof(buf));

		// we need to load the sound
		err = soundfile_open(&fil,fn,subfn);
		if (err != jfauderr_noerror) {
			debuglog(DEBUGLVL_TRACE,"soundfile_open() failed");
			goto cleanup;
		}

		if (!fil->sf) {
			debuglog(DEBUGLVL_TRACE,"file not recognised");
			soundfile_close(fil);
			goto cleanup;
		}

		if (fil->sf->type == type_wave) {
			buf.type = buffertype_pcm;

			err = filereaders_wave_getplaytime(fil, &playlen);
			if (err != jfauderr_noerror && err != jfauderr_notimpl) {
				debuglog(DEBUGLVL_TRACE,"failed getting waveform play time");
				soundfile_close(fil);
				goto cleanup;
			}

			if (err == jfauderr_notimpl) stream = 1;
			else stream = shouldstream(playlen);

			if (!stream) {
				err = filereaders_wave_readallsamples(fil, &buf.buf.pcm);
				if (err != jfauderr_noerror) {
					debuglog(DEBUGLVL_TRACE,"failed reading all samples");
					soundfile_close(fil);
					goto cleanup;
				}
				soundfile_close(fil);

				if (id.a.flags & fileidentflag_fxstertomono)
					if ((err = fx_stertomono(&buf.buf.pcm)) != jfauderr_noerror) {
						debuglog(DEBUGLVL_TRACE,"stereo-to-mono filter failed");
						pcmbuffer_free(&buf.buf.pcm);
						goto cleanup;
					}
				if (id.a.flags & fileidentflag_fxnearest)
					if ((err = fx_nearest(&buf.buf.pcm)) != jfauderr_noerror) {
						debuglog(DEBUGLVL_TRACE,"nearest filter failed");
						pcmbuffer_free(&buf.buf.pcm);
						goto cleanup;
					}
				if (id.a.flags & fileidentflag_doublespeed)
					buf.buf.pcm.samplerate <<= 1;

				err = openal_preparebuffer(&buf.buf.pcm);
				if (err != jfauderr_noerror) {
					debuglog(DEBUGLVL_TRACE,"openal failed preparing buffer");
					pcmbuffer_free(&buf.buf.pcm);
					goto cleanup;
				}
			}
		} else if (fil->sf->type == type_midi) {
			buf.type = buffertype_midi;

			err = filereaders_midi_getdata(fil, &buf.buf.midi);
			if (err != jfauderr_noerror) {
				debuglog(DEBUGLVL_TRACE,"failed reading MIDI data");
				soundfile_close(fil);
				goto cleanup;
			}
			soundfile_close(fil);
		}

		soundtype = buf.type;
		err = buffer_manage(id, &buf);
		if (err != jfauderr_noerror) {
			debuglog(DEBUGLVL_TRACE,"failed submitting buffer for management");
			goto cleanup;
		}
	}

	if (soundtype == -2) {	// CD audio special case
		int starttrack = -1, endtrack = -1;
		char *p = NULL;

		if (subfn) {
			starttrack = (int)strtol(subfn,&p,10);
			if (p && *p == '-') {
				endtrack = (int)strtol(p+1,NULL,10);
			} else endtrack = starttrack;
		}

		err = channel_alloccda(&chan, starttrack,endtrack);
	} else if (soundtype == buffertype_pcm) {
		if (!stream) err = channel_allocwave(&chan, id, priority);
		else {
			fil->ident.a.flags = id.a.flags;
			err = channel_allocwavestream(&chan, fil, priority);
		}
	} else if (soundtype == buffertype_midi) {
		if (mode != 1) {
			debuglog(DEBUGLVL_TRACE,"tried to play a MIDI file as a sound");
			err = jfauderr_notcompatible;
			goto cleanup;
		}

		err = channel_allocmidi(&chan, id);
	}
	
	if (err != jfauderr_noerror) {
		debuglog(DEBUGLVL_TRACE,jsprintf("failed allocating channel (soundtype=%d stream=%d)",soundtype,stream));
		goto cleanup;
	}

	if (danprops > 0) {
		err = channel_setprops(chan, danprops, daprops);
		if (err != jfauderr_noerror) {
			debuglog(DEBUGLVL_TRACE,"failed setting channel properties");
			channel_release(chan);
			goto cleanup;
		}
	}

	*handle = (int)chan;
	err = channel_setplaymode(chan, initial);

	if (err != jfauderr_noerror)
		debuglog(DEBUGLVL_TRACE,"failed setting play mode");

cleanup:
	free(daprops);
	return err;
}

jfauderr JFAud_playsound(int *handle, const char *fn, const char *subfn, int priority, JFAudPlayMode initial,
		int nprops, JFAudProp *props)
{
	if (!handle || !fn) return jfauderr_genericerror;
	if (priority <= ReservedPriority) return jfauderr_genericerror;
	return playsound(handle,fn,subfn,priority,initial,nprops,props, 0);
}

jfauderr JFAud_playrawsound(int *handle, const void *data, unsigned int dataleng,
		JFAudRawFormat *raw, int priority, JFAudPlayMode initial)
{
	return jfauderr_notimpl;
}

jfauderr JFAud_playmusic(const char *fn, const char *subfn, JFAudPlayMode initial, int nprops, JFAudProp *props)
{
	JFAudProp daprops[8];
	int i;

	// set up the defaults
	daprops[0].prop = JFAudProp_gain;     daprops[0].val.f = 1.0;
	daprops[1].prop = JFAudProp_pitch;    daprops[1].val.f = 1.0;
	daprops[2].prop = JFAudProp_loop;     daprops[2].val.i = 1;
	daprops[3].prop = JFAudProp_rolloff;  daprops[3].val.f = 0.0;
	daprops[4] = mkpropv(JFAudProp_position,0.0,0.0,0.0);
	daprops[5] = mkpropv(JFAudProp_velocity,0.0,0.0,0.0);
	daprops[6].prop = JFAudProp_posrel;   daprops[6].val.i = 0;
	daprops[7].prop = JFAudProp_fx;       daprops[7].val.i = 0;

	for (i=0;i<nprops;i++) {
		switch (props[i].prop) {
			case JFAudProp_gain: daprops[0].val.f = props[i].val.f; break;
			case JFAudProp_loop: daprops[2].val.i = props[i].val.i; break;
			case JFAudProp_fx:   daprops[7].val.i = props[i].val.i; break;
			default: break;
		}
	}
	
	if (musicchan >= 0) channel_release(musicchan);
	musicchan = -1;
	return playsound(&musicchan,fn,subfn,ReservedPriority,initial,8,daprops, 1);
}

jfauderr JFAud_killsound(int handle)
{
	return channel_release(handle);
}

jfauderr JFAud_setlistenerprops(int nprops, JFAudProp *props)
{
	int i;
	jfauderr err = jfauderr_noerror;
	int danprops;
	JFAudProp *daprops;

	if (nprops <= 0) return jfauderr_noerror;
	if (!props) return jfauderr_genericerror;
	
	daprops = (JFAudProp *)calloc(nprops,sizeof(JFAudProp));
	if (!daprops) return jfauderr_outofmemory;
	danprops = 0;

	for (i=0;i<nprops;i++) {
		daprops[danprops] = props[i];
		switch (props[i].prop) {
			case JFAudProp_gain:
			case JFAudProp_position:
			case JFAudProp_velocity:
			case JFAudProp_orient:
				danprops++;
				break;
			default: break;
		}
	}

	if (danprops > 0) err = openal_setlistenerprops(danprops, daprops);
	free(daprops);
	return err;
}

jfauderr JFAud_setsoundprops(int handle, int nprops, JFAudProp *props)
{
	int i;
	jfauderr err = jfauderr_noerror;
	int danprops;
	JFAudProp *daprops;

	if (nprops <= 0) return jfauderr_noerror;
	if (!props) return jfauderr_genericerror;

	daprops = (JFAudProp *)calloc(nprops,sizeof(JFAudProp));
	if (!daprops) return jfauderr_outofmemory;
	danprops = 0;

	for (i=0;i<nprops;i++) {
		daprops[danprops] = props[i];
		switch (props[i].prop) {
			case JFAudProp_gain:
			case JFAudProp_pitch:
			case JFAudProp_loop:
			case JFAudProp_rolloff:
			case JFAudProp_position:
			case JFAudProp_velocity:
			case JFAudProp_posrel:
			case JFAudProp_mindist:
			//case JFAudProp_fx:	// this can only be set at sound start time
				danprops++;
				break;
			default: break;
		}
	}

	if (danprops > 0) err = channel_setprops((jfaudchan)handle, danprops, daprops);
	free(daprops);
	return err;
}

jfauderr JFAud_setmusicprops(int nprops, JFAudProp *props)
{
	int have[2] = { 0,0 };	// gain and loop (fx can't be changed on the fly)
	JFAudProp daprops[2];
	int i;

	if (musicchan < 0) return jfauderr_genericerror;	// nothing playing

	for (i=0;i<nprops;i++) {
		switch (props[i].prop) {
			case JFAudProp_gain: have[0] = 1; daprops[0] = props[i]; break;
			case JFAudProp_loop: have[1] = 1; daprops[1] = props[i]; break;
			default: break;
		}
	}

	if (!have[0] && !have[1]) return jfauderr_noerror;

	if (!have[0] && have[1]) {
		daprops[0] = daprops[1];
		have[0] = 1;
		have[1] = 0;
	}

	return channel_setprops(musicchan, have[0]+have[1], daprops);
}

jfauderr JFAud_getsoundplaymode(int handle, JFAudPlayMode *mode)
{
	return channel_getplaymode((jfaudchan)handle, mode);
}

jfauderr JFAud_setsoundplaymode(int handle, JFAudPlayMode mode)
{
	return channel_setplaymode((jfaudchan)handle, mode);
}

jfauderr JFAud_getmusicplaymode(JFAudPlayMode *mode)
{
	if (musicchan < 0) { *mode = JFAudPlayMode_stopped; return jfauderr_noerror; }
	return channel_getplaymode(musicchan, mode);
}

jfauderr JFAud_setmusicplaymode(JFAudPlayMode mode)
{
	if (musicchan < 0) return jfauderr_noerror;
	return channel_setplaymode(musicchan, mode);
}

jfauderr JFAud_update(void)
{
	if (conf.synchronous) {
		channel_update();
		buffer_update();
		return jfauderr_noerror;
	} else {
		return jfauderr_noerror;
	}
}


const char *JFAud_errstr(jfauderr err)
{
	switch (err) {
		case jfauderr_noerror: return "No error";
		case jfauderr_genericerror: return "Generic error";
		case jfauderr_alreadyinited: return "Already initialised";
		case jfauderr_notimpl: return "Not implemented";
		case jfauderr_disabled: return "File reader disabled";
		case jfauderr_openerror: return "File open error";
		case jfauderr_notrecognised: return "File type not recognised";
		case jfauderr_notcompatible: return "File type recognised but format incompatible";
		case jfauderr_filedamaged: return "File is damaged";
		case jfauderr_notloaded: return "Buffer not loaded";
		case jfauderr_notprepared: return "Buffer not prepared";
		case jfauderr_buffull: return "Buffer is full";
		case jfauderr_badhandle: return "Invalid handle";
		case jfauderr_nofreechans: return "No free channels available";

		case jfauderr_outofmemory: return "Out of memory";
		case jfauderr_deverror: return "Device error";
		case jfauderr_ioerror: return "IO error";
		case jfauderr_notinited: return "Not initialised";
		case jfauderr_initfailed: return "Initialisation failed";

		default: return "Unknown error";
	}
}

void logmsg(const char *s) { conf.cb_logmsg(s); }

#include <stdarg.h>
#ifdef _MSC_VER
#define vsnprintf _vsnprintf
#endif
static char sprintfbuf[1024];
char *jsprintf(const char *fmt, ...)
{
	va_list va;
	va_start(va,fmt);
	vsnprintf(sprintfbuf,1024,fmt,va);
	va_end(va);
	return sprintfbuf;
}
void debugmsg(int level, const char *s, const char *fn, const char *func, int line)
{
	char msg[1024];
	if (debuglevel < level) return;
	if (func) snprintf(msg,1024,"(%d)%s %s:%d %s",level,fn,func,line,s);
	else snprintf(msg,1024,"(%d)%s:%d %s",level,fn,line,s);
	logmsg(msg);
}

