#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "internal.h"

typedef struct {
	MidiFile_defn

	unsigned format, numtracks, ppqn;
} MThd;

static unsigned long getbignum(unsigned char *ptr, int size)
{
        unsigned long val = 0;
        if (size > 4) size = 4;
        while (size--) {
                val <<= 8;
                val |= *(ptr++);
        }
        return val;
}


// SoundFileReader functions

static jfauderr init(void)
{
	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	MThd mthd;
	unsigned char headerbuf[8];
	unsigned long chunksz;
	
	if (!file) return jfauderr_genericerror;
	if (file->sf) return jfauderr_genericerror;	// already been identified

	sysf_seek(&file->fh, 0, SEEK_SET);

	if (sysf_read(&file->fh, headerbuf, 4*2) != 4*2) return jfauderr_notrecognised;

	if (headerbuf[0] != 'M' || headerbuf[1] != 'T' || headerbuf[2] != 'h' || headerbuf[3] != 'd')
		return jfauderr_notrecognised;

	if ((chunksz=getbignum(&headerbuf[4], 4)) < 6) return jfauderr_filedamaged;
	if (sysf_read(&file->fh, headerbuf, 2*3) != 2*3) return jfauderr_filedamaged;

	mthd.format = getbignum(&headerbuf[0], 2);
	mthd.numtracks = getbignum(&headerbuf[2], 2);
	mthd.ppqn = getbignum(&headerbuf[4], 2);

	debuglog(DEBUGLVL_GENERAL,jsprintf("midiseq: Format: %d, Tracks: %d, Ticks/quarter note: %d",
				mthd.format,mthd.numtracks,mthd.ppqn));
	
	mthd.type = type_midi;
	
	file->sf = (SoundFile *)malloc(sizeof(MThd));
	if (!file->sf) return jfauderr_outofmemory;	// out of memory

	memcpy(file->sf, &mthd, sizeof(MThd));

	return jfauderr_noerror;
}

static jfauderr unload(File *file UNUSED, char *noclose)
{
	// caller cleans up file->sf
	*noclose = 0;

	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file UNUSED, JFAudMetaInfo what UNUSED, char *buf UNUSED, int *len UNUSED)
{
	return jfauderr_notimpl;
}

static jfauderr getdata(File *file, MidiBuffer *buf)
{
	unsigned i, tracklen;
	jfauderr err;
	unsigned char headerbuf[8];
	MThd *mthd = (MThd*)file->sf;

	err = midibuffer_allocate(buf, mthd->numtracks);
	if (err != jfauderr_noerror) return err;

	buf->numtracks = mthd->numtracks;
	buf->ppqn      = mthd->ppqn;
        
	for (i=0; i<mthd->numtracks; i++) {
		if (sysf_read(&file->fh, headerbuf, 4*2) != 4*2) {
			debuglog(DEBUGLVL_GENERAL,"Truncated track.");
			midibuffer_free(buf);
			return jfauderr_filedamaged;
		}

		if (headerbuf[0] != 'M' || headerbuf[1] != 'T' || headerbuf[2] != 'r' || headerbuf[3] != 'k') {
			debuglog(DEBUGLVL_GENERAL,"Track header invalid.");
			midibuffer_free(buf);
			return jfauderr_filedamaged;
		}

		tracklen = getbignum(&headerbuf[4], 4);
		err = midibuffer_allocatetrack(buf, i, tracklen);
		if (err != jfauderr_noerror) {
			midibuffer_free(buf);
			return err;
		}

		debuglog(DEBUGLVL_GENERAL,jsprintf("Track %d: size=%d",i,tracklen));

		if (sysf_read(&file->fh, buf->tracks[i], tracklen) != (int)tracklen) {
			debuglog(DEBUGLVL_GENERAL,"Track data truncated.");
			midibuffer_free(buf);
			return jfauderr_filedamaged;
		}
	}

	return jfauderr_noerror;
}


MidiFileReader mf_mid = {
	"Standard MIDI",
	init,
	uninit,

	type_midi,
	test,
	unload,
	getmetainfo,

	getdata
};


