#include "internal.h"
#include "midiseq.h"

#include <stdlib.h>
#include <string.h>

enum { IgnoreTrack = 1, IgnorePrograms = 2, IgnoreVolumes = 4, TrackDisabled = 8 };

static const int cmdlens[] = {
        2 /*noteoff 0x80*/, 2 /*noteon 0x90*/, 2 /*aftertouch 0xA0*/, 2 /*control 0xB0*/,
        1 /*program 0xC0*/, 1 /*channelaftertouch 0xD0*/, 2 /*pitchwheel 0xE0*/, 0 /*syscommon 0xF0*/
};

static unsigned long getbignum(unsigned char *ptr, int size)
{
        unsigned long val = 0;
        if (size > 4) size = 4;
        while (size--) {
                val <<= 8;
                val |= *(ptr++);
        }
        return val;
}


jfauderr midifile_loadfile(MidiBuffer *midi, File *fh)
{
	unsigned long chunksz;
	unsigned char headerbuf[2*8];
	jfauderr err;

	unsigned i, format, numtracks, ppqn, tracklen;

	if (sysf_read(&fh->fh, headerbuf, 4*2) != 4*2) return jfauderr_notrecognised;

	if (headerbuf[0] != 'M' || headerbuf[1] != 'T' || headerbuf[2] != 'h' || headerbuf[3] != 'd') {
		debuglog(DEBUGLVL_GENERAL,"midiseq: File doesn't look like MIDI.");
		return jfauderr_notrecognised;
	}

	if ((chunksz=getbignum(&headerbuf[4], 4)) < 6) {
		debuglog(DEBUGLVL_GENERAL,"midiseq: Header is too short.");
		return jfauderr_filedamaged;
	}

	if (sysf_read(&fh->fh, headerbuf, 2*3) != 2*3) {
		debuglog(DEBUGLVL_GENERAL,"midiseq: Truncated header.");
		return jfauderr_filedamaged;
	}

	format = getbignum(&headerbuf[0], 2);
	numtracks = getbignum(&headerbuf[2], 2);
	ppqn = getbignum(&headerbuf[4], 2);

	debuglog(DEBUGLVL_GENERAL,jsprintf("midiseq: Format: %d, Tracks: %d, Ticks/quarter note: %d",format,numtracks,ppqn));

	err = midibuffer_allocate(midi, numtracks);
	if (err != jfauderr_noerror) return err;
        
	for (i=0; i<numtracks; i++) {
		if (sysf_read(&fh->fh, headerbuf, 4*2) != 4*2) {
			debuglog(DEBUGLVL_GENERAL,"midiseq: Truncated track.");
			midibuffer_free(midi);
			return jfauderr_filedamaged;
		}

		if (headerbuf[0] != 'M' || headerbuf[1] != 'T' || headerbuf[2] != 'r' || headerbuf[3] != 'k') {
			debuglog(DEBUGLVL_GENERAL,"midiseq: Track header invalid.");
			midibuffer_free(midi);
			return jfauderr_filedamaged;
		}

		tracklen = getbignum(&headerbuf[4], 4);
		err = midibuffer_allocatetrack(midi, i, tracklen);
		if (err != jfauderr_noerror) {
			midibuffer_free(midi);
			return err;
		}

		debuglog(DEBUGLVL_GENERAL,jsprintf("midiseq: Track %d: size=%d",i,tracklen));

		if (sysf_read(&fh->fh, midi->tracks[i], tracklen) != (int)tracklen) {
			debuglog(DEBUGLVL_GENERAL,"midiseq: Track data truncated.");
			midibuffer_free(midi);
			return jfauderr_filedamaged;
		}
	}

	return jfauderr_noerror;
}



static inline void setflag(MidiSequencer *seq, char flag) { seq->seqtrack[seq->curtrack].trackflags |= flag; }
static inline void clrflag(MidiSequencer *seq, char flag) { seq->seqtrack[seq->curtrack].trackflags &= ~flag; }
static inline char testflag(MidiSequencer *seq, char flag) { return (seq->seqtrack[seq->curtrack].trackflags & flag); }

static inline int eot(MidiSequencer *seq) {
	if (seq->curtrack >= seq->midifile->numtracks) return 1;
	if ((long)seq->seqtrack[seq->curtrack].trackpos - (long)seq->midifile->tracks[seq->curtrack] >=
			(long)seq->midifile->tracklength[seq->curtrack]) return 1;
	return 0;
}
static inline short getbyte(MidiSequencer *seq) {
	if (eot(seq)) return -1;
	return *(seq->seqtrack[seq->curtrack].trackpos++);
}
static inline short peekbyte(MidiSequencer *seq, unsigned short ahead) {
	if (seq->curtrack >= seq->midifile->numtracks) return 1;
	if ((long)seq->seqtrack[seq->curtrack].trackpos+ahead-(long)seq->midifile->tracks[seq->curtrack] >=
			(long)seq->midifile->tracklength[seq->curtrack]) return -1;
	return seq->seqtrack[seq->curtrack].trackpos[ahead];
}
static inline unsigned long getpos(MidiSequencer *seq) { return (long)seq->seqtrack[seq->curtrack].trackpos - 
								(long)seq->midifile->tracks[seq->curtrack]; }
static inline unsigned char const *getptr(MidiSequencer *seq) { return seq->seqtrack[seq->curtrack].trackpos; }
static inline void setpos(MidiSequencer *seq, unsigned long pos) { seq->seqtrack[seq->curtrack].trackpos =
						(unsigned char*)((long)seq->midifile->tracks[seq->curtrack] + pos); }
                        
static inline void selecttrack(MidiSequencer *seq, int track) { seq->curtrack = track; }
static inline void rewind(MidiSequencer *seq) { seq->seqtrack[seq->curtrack].trackpos = seq->midifile->tracks[seq->curtrack]; }

static unsigned long getmidinum(MidiSequencer *seq)
{
        long val = 0;
        unsigned char by;
        do {
                by = getbyte(seq);
                val <<= 7;
                val |= (by&0x7f);
        } while (by&0x80);
        return val;
}

static void midiseq_scan(MidiSequencer *seq, unsigned char device)
{
	unsigned i;
	char foundincl;
	long deltatime;
	unsigned char command, parm1, parm2;

	for (i=0; i<seq->midifile->numtracks; i++) {
		foundincl = 0;
		command = 0;
		selecttrack(seq, i);
		rewind(seq);
		clrflag(seq, 255);
		while (!eot(seq)) {
			deltatime = getmidinum(seq);
			if (peekbyte(seq, 0) & 0x80) command = getbyte(seq);
			else if (!command)
				while (!((command = getbyte(seq)) & 0x80)) ;

			if ((command&0xF0) == 0xB0) {	// controller change
				parm1 = getbyte(seq);
				parm2 = getbyte(seq);
				switch (parm1) {
					case 110:	// EMIDI include
						if (parm2 == 127 || parm2 == device) clrflag(seq, IgnoreTrack);
						else if (!foundincl) setflag(seq, IgnoreTrack);
						foundincl = 1;
						break;
                                                
					case 111:	// EMIDI exclude
						if (parm2 == 127 || parm2 == device)
							setflag(seq, IgnoreTrack);
						break;

					case 112:	// EMIDI program
						setflag(seq, IgnorePrograms);
						break;

					case 113:	// EMIDI volume
						setflag(seq, IgnoreVolumes);
						break;
				}
			} else if ((command&0xF0) == 0xF0) {	// meta event
				switch (command) {
					case 0xF0:
						while (getbyte(seq) != 0xF7) ;
						break;
					case 0xFF:
						getbyte(seq);
						for (parm2=getbyte(seq); parm2>0; parm2--) getbyte(seq);
						break;
					case 0xF2:
						getbyte(seq);
					case 0xF1: case 0xF3:
						getbyte(seq);
					default:
						break;
				}
				command = 0;
			} else {
				for (parm2=cmdlens[ (command>>4) & 7 ]; parm2>0; parm2--)
					getbyte(seq);
			}
		}
	}

	for (i=0; i<seq->midifile->numtracks; i++) {
		selecttrack(seq,i);
		debuglog(DEBUGLVL_WORDY,jsprintf("track %d is %s",i,testflag(seq,IgnoreTrack)?"ignored":"enabled"));
	}
}

static int midiseq_fetchevt(MidiSequencer *seq, int track)
{
	int cmd,len;
        
	selecttrack(seq, track);
	if (eot(seq)) return 1;

	seq->seqtrack[track].delta = getmidinum(seq);
	if (peekbyte(seq, 0) & 0x80) seq->seqtrack[track].command = getbyte(seq);
	cmd = seq->seqtrack[track].command;
	len = cmdlens[ (cmd>>4) & 7 ];
	seq->seqtrack[track].evtptr = getptr(seq);
	seq->seqtrack[track].evtlen = len;

	if (len == 0) {	// system common event
		switch (cmd) {
			case 0xF0:
				// sysex
				do seq->seqtrack[track].evtlen++; while (getbyte(seq) != 0xF7);
				break;
			case 0xFF:
				// meta
				cmd = getbyte(seq);
				len = getbyte(seq);
				seq->seqtrack[track].evtlen += len + 2;
				while (len--) getbyte(seq);
				break;
			case 0xF2:
				// two-byte messages
				seq->seqtrack[track].evtlen++; getbyte(seq);
			case 0xF1:
			case 0xF3:
				// one-byte messages
				seq->seqtrack[track].evtlen++; getbyte(seq);
			default:
				// command-byte-only messages
				break;
		}
	} else {
		while (len--) getbyte(seq);
	}

	return 0;
}




jfauderr midiseq_new(MidiSequencer *seq, MidiBuffer *midi)
{
	seq->seqtrack = (MidiSequencerTrackdef *)calloc(midi->numtracks, sizeof(MidiSequencerTrackdef));
	if (!seq->seqtrack) return jfauderr_outofmemory;
	
	seq->midifile = midi;
	seq->lastcommand = 0;
	seq->infiniteloop = 0;

	return jfauderr_noerror;
}

jfauderr midiseq_free(MidiSequencer *seq)
{
	if (seq->seqtrack) free(seq->seqtrack);
	return jfauderr_noerror;
}

jfauderr midiseq_setdevice(MidiSequencer *seq, int dev)
{
	unsigned i;

	if ((unsigned)dev > 127) dev = 127;
        
	memset(seq->seqtrack, 0, sizeof(MidiSequencerTrackdef) * seq->midifile->numtracks);
	seq->lastcommand = 0;

	midiseq_scan(seq, dev);

        for (i=0;i<seq->midifile->numtracks;i++) {
                selecttrack(seq, i);
                if (testflag(seq, IgnoreTrack)) { setflag(seq, TrackDisabled); continue; }
                rewind(seq);
                if (midiseq_fetchevt(seq, i)) setflag(seq, TrackDisabled);
        }

        return jfauderr_noerror;
	
}

jfauderr midiseq_setloop(MidiSequencer *seq, int loop)
{
	seq->infiniteloop = loop;
	return jfauderr_noerror;
}

int midiseq_getevent(MidiSequencer *seq, unsigned *delta, unsigned char *command, unsigned *len, unsigned char const **evtdata)
{
	// returns 0 if a message was returned, 1 if the end of the song was reached, or -1 on error

	unsigned long deltaacc = 0, mindelta = 0xfffffff;
	int mindeltatrk = -1, j;
	unsigned i;
	char gotevent = 0;

again:
	// find the track with the nearest event
	mindelta = 0xffffffff;
	mindeltatrk = -1;
	for (i=0;i<seq->midifile->numtracks;i++) {
		if (seq->seqtrack[i].trackflags & TrackDisabled) continue;
		if (seq->seqtrack[i].delta >= mindelta) continue;
		mindelta = seq->seqtrack[i].delta;
		mindeltatrk = i;
	}
	if (mindeltatrk < 0) {
		if (!seq->infiniteloop) return 1;
		for (i=0;i<seq->midifile->numtracks;i++) {
			selecttrack(seq, i);
			if (testflag(seq, IgnoreTrack)) { setflag(seq, TrackDisabled); continue; }
			clrflag(seq, TrackDisabled);
			rewind(seq);
			if (midiseq_fetchevt(seq, i)) setflag(seq, TrackDisabled);
		}
		goto again;
	}

	// check out the event
	selecttrack(seq, mindeltatrk);
	switch (seq->seqtrack[mindeltatrk].command & 0xF0) {
		case 0xC0:	// program
			if (testflag(seq, IgnorePrograms))
				break;
		case 0x80:	// note off
		case 0x90:	// note on
		case 0xA0:	// aftertouch
		case 0xD0:	// channel aftertouch
		case 0xE0:	// pitch wheel
			gotevent = 1;
			break;
		case 0xF0:	// system common event
			// we keep tempo meta events so they can be interpreted, but all other metas get swallowed
			if (seq->seqtrack[mindeltatrk].command == 0xFF && seq->seqtrack[mindeltatrk].evtptr[0] == 0x51)
				gotevent = 1;
			else if (seq->seqtrack[mindeltatrk].command == 0xFF &&
				 seq->seqtrack[mindeltatrk].evtptr[0] >= 1 &&
				 seq->seqtrack[mindeltatrk].evtptr[0] <= 7) {
				char msg[256], *type;
				memset(msg,0,256);
				memcpy(msg, &seq->seqtrack[mindeltatrk].evtptr[2], seq->seqtrack[mindeltatrk].evtptr[1]);
				switch (seq->seqtrack[mindeltatrk].evtptr[0]) {
					case 1: type = "text"; break;
					case 2: type = "copyright"; break;
					case 3: type = "track name"; break;
					case 4: type = "instrument name"; break;
					case 5: type = "lyric"; break;
					case 6: type = "marker"; break;
					case 7: type = "cue point"; break;
					default: type = "?"; break;
				}
				debuglog(DEBUGLVL_WORDY,jsprintf("%s event: %s",type,msg));
			}
			else if (seq->seqtrack[mindeltatrk].command == 0xFF && seq->seqtrack[mindeltatrk].evtptr[0] == 0x2F)
				setflag(seq, TrackDisabled);
			break;
		case 0xB0:	// control
			switch (seq->seqtrack[mindeltatrk].evtptr[0]) {
				case 110:	// EMIDI include
				case 111:	// EMIDI exclude
				case 114:	// EMIDI context start
				case 115:	// EMIDI context end
					break;
				case 113:	// EMIDI volume
					seq->evtbuf[0] = 7;		// cook up a normal volume change event
					seq->evtbuf[1] = seq->seqtrack[mindeltatrk].evtptr[1];
					*command = seq->seqtrack[mindeltatrk].command;
					*evtdata = seq->evtbuf;
					*len     = 2;
					gotevent = 2;
					break;
				case 112:	// EMIDI program
					*command  = 0xC0 | (seq->seqtrack[mindeltatrk].command & 15);	// cook up a normal
					seq->evtbuf[0] = seq->seqtrack[mindeltatrk].evtptr[1] & 0x7F;	// program change event
					*evtdata = seq->evtbuf;
					*len     = 1;
					gotevent = 2;
					break;
				case 116:	// EMIDI loop begin
				case 118:	// EMIDI song loop begin
					if (seq->seqtrack[mindeltatrk].evtptr[1] == 0) {
						j = -1;
					} else {
						j = seq->seqtrack[mindeltatrk].evtptr[1];
						debuglog(DEBUGLVL_WORDY,jsprintf("Track %d: Entering %d-cycle loop",mindeltatrk,j));
					}

					if (seq->seqtrack[mindeltatrk].evtptr[0] == 118) {
						// song-wide loop
						for (i=0;i<seq->midifile->numtracks;i++) {
							seq->seqtrack[i].loopcount = j;
							selecttrack(seq, i);
							seq->seqtrack[i].loopstart = getpos(seq);
						}
					} else {
						seq->seqtrack[mindeltatrk].loopcount = j;
						seq->seqtrack[mindeltatrk].loopstart = getpos(seq);
					}
					break;
				case 117:	// EMIDI loop end
				case 119:	// EMIDI song loop end
					if (seq->seqtrack[mindeltatrk].evtptr[1] != 127) {
						break;
					}
					if (!seq->infiniteloop && seq->seqtrack[mindeltatrk].loopcount < 0) {
						break;
					}

					if (seq->seqtrack[mindeltatrk].evtptr[0] == 118) {
						// song-wide loop
						for (i=0;i<seq->midifile->numtracks;i++) {
							selecttrack(seq, i);
							seq->seqtrack[i].loopend = getpos(seq);
						}
						selecttrack(seq, mindeltatrk);
					}

					if (seq->seqtrack[mindeltatrk].loopcount) {
						if (seq->seqtrack[mindeltatrk].loopcount > 0) {
							seq->seqtrack[mindeltatrk].loopcount--;
							debuglog(DEBUGLVL_WORDY,jsprintf("Track %d: %d to go...",mindeltatrk,seq->seqtrack[mindeltatrk].loopcount));
						}
						setpos(seq, seq->seqtrack[mindeltatrk].loopstart);
					}
					break;
				case 7:		// volume change
					if (testflag(seq, IgnoreVolumes))
						break;
				default:
					gotevent = 1;
					break;
			}
			break;
	}

	if (gotevent) {
		if (gotevent == 1) {
			*command = seq->seqtrack[mindeltatrk].command;
			*evtdata = seq->seqtrack[mindeltatrk].evtptr;
			*len     = seq->seqtrack[mindeltatrk].evtlen;
		}
		seq->lastcommand = *command;
		*delta = deltaacc + mindelta;
	}
        
	// move all events forward and get a new one for the track we just ate
	for (i=0;i<seq->midifile->numtracks;i++) {
		if (seq->seqtrack[i].trackflags & TrackDisabled) continue;
		if (i == (unsigned)mindeltatrk) {
			if (midiseq_fetchevt(seq, i)) {
				seq->seqtrack[i].trackflags |= TrackDisabled;
			}
		} else {
			seq->seqtrack[i].delta -= mindelta;
		}
	}

	if (gotevent) {
		return 0;
	} else {
		deltaacc += mindelta;
		goto again;
	}

	return -1;
}

