#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "internal.h"

#if (WITH_MOD == 0)
static jfauderr init(void) { return jfauderr_disabled; }
static jfauderr uninit(void) { return jfauderr_disabled; }
static jfauderr test(File *file UNUSED) { return jfauderr_disabled; }
static jfauderr unload(File *file UNUSED, char *noclose UNUSED) { return jfauderr_disabled; }
static jfauderr getmetainfo(File *file UNUSED, char what UNUSED, char *buf UNUSED, int len UNUSED) { return jfauderr_disabled; }
static jfauderr readsamples(File *file UNUSED, int n UNUSED, PcmBuffer *buf UNUSED, char loop UNUSED) { return jfauderr_disabled; }
static jfauderr readallsamples(File *file UNUSED, PcmBuffer *buf UNUSED) { return jfauderr_disabled; }
static jfauderr getplaytime(File *file UNUSED, float *len UNUSED) { return jfauderr_disabled; }
#else

#include "modplug.h"

// the largest XM in my collection is aws_fan2.xm at 5368865 bytes
// and this is so people don't get clever and try giving us a 200MB
// file or something, knowing the last resort test (currently this MOD
// loader) will load it all into memory
#define LARGESTMOD (1048576*6)

#if (DYNAMIC_MOD == 1)

#ifdef _WIN32
#ifdef PACKED
#undef PACKED
#endif
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define DLLNAME "libmodplug.dll"
#else

#include <dlfcn.h>
#define DLLNAME "libmodplug.so"
#endif


static struct {
#ifdef _WIN32
        HMODULE dll;
#else
        void *dll;
#endif

	ModPlugFile* (__cdecl *ModPlug_Load)(const void* data, int size);
	void (__cdecl *ModPlug_Unload)(ModPlugFile* file);
	int  (__cdecl *ModPlug_Read)(ModPlugFile* file, void* buffer, int size);
	const char* (__cdecl *ModPlug_GetName)(ModPlugFile* file);
	int (__cdecl *ModPlug_GetLength)(ModPlugFile* file);
	void (__cdecl *ModPlug_Seek)(ModPlugFile* file, int millisecond);
	void (__cdecl *ModPlug_GetSettings)(ModPlug_Settings* settings);
	void (__cdecl *ModPlug_SetSettings)(const ModPlug_Settings* settings);
} libmodplug = {NULL, NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

#define ModPlug_Load libmodplug.ModPlug_Load
#define ModPlug_Unload libmodplug.ModPlug_Unload
#define ModPlug_Read libmodplug.ModPlug_Read
#define ModPlug_GetName libmodplug.ModPlug_GetName
#define ModPlug_GetLength libmodplug.ModPlug_GetLength
#define ModPlug_Seek libmodplug.ModPlug_Seek
#define ModPlug_GetSettings libmodplug.ModPlug_GetSettings
#define ModPlug_SetSettings libmodplug.ModPlug_SetSettings

#define testifloaded() { if (libmodplug.dll == NULL) return jfauderr_notinited; }
#else

#define testifloaded() { }
#endif


typedef struct {
	WaveformFile_defn

	ModPlugFile *mod;
} ModPlug;


// SoundFileReader functions

static jfauderr init(void)
{
	ModPlug_Settings set;

#if (DYNAMIC_MOD == 1)
        int er=0;

#ifdef _WIN32
        void *sym;
        libmodplug.dll = LoadLibrary(DLLNAME);

#define gpa(d,x) sym = GetProcAddress(d,x); { \
                if (!sym) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#else
        libmodplug.dll = dlopen(DLLNAME, RTLD_GLOBAL|RTLD_NOW);
        dlerror();

#define gpa(d,x) dlsym(d,x); { \
                const char *s; \
                if ((s = dlerror())) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#endif

        if (!libmodplug.dll) return jfauderr_disabled;
        
        ModPlug_Load = (void*)gpa(libmodplug.dll, "ModPlug_Load");
        ModPlug_Unload = (void*)gpa(libmodplug.dll, "ModPlug_Unload");
        ModPlug_Read = (void*)gpa(libmodplug.dll, "ModPlug_Read");
        ModPlug_GetName = (void*)gpa(libmodplug.dll, "ModPlug_GetName");
        ModPlug_GetLength = (void*)gpa(libmodplug.dll, "ModPlug_GetLength");
        ModPlug_Seek = (void*)gpa(libmodplug.dll, "ModPlug_Seek");
        ModPlug_GetSettings = (void*)gpa(libmodplug.dll, "ModPlug_GetSettings");
        ModPlug_SetSettings = (void*)gpa(libmodplug.dll, "ModPlug_SetSettings");

        if (er) {
#ifdef _WIN32
                FreeLibrary(libmodplug.dll);
#else
                dlclose(libmodplug.dll);
#endif

                libmodplug.dll = NULL;
                debuglog(DEBUGLVL_GENERAL,"Failed fetching required symbols from " DLLNAME ".");
                return jfauderr_initfailed;
        }
#endif	

	memset(&set,0,sizeof(set));
	set.mFlags          = MODPLUG_ENABLE_OVERSAMPLING;
	set.mChannels       = 2;
	set.mBits           = 16;
	set.mFrequency      = min(conf.samplerate,44100);
	set.mResamplingMode = MODPLUG_RESAMPLE_SPLINE;
	set.mLoopCount      = -1;	// FIXME: this will be interesting if we don't wish to loop the song...

	ModPlug_SetSettings(&set);
	
	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
#if (DYNAMIC_MOD == 1)
        if (libmodplug.dll) {
#ifdef _WIN32
                FreeLibrary(libmodplug.dll);
#else
                dlclose(libmodplug.dll);
#endif

                libmodplug.dll = NULL;
        }
#endif
	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	ModPlug mod;
	void *data;
	int length;

	testifloaded();

	if (!file) return jfauderr_genericerror;
	if (file->sf) return jfauderr_genericerror;	// already been identified

	mod.type = type_wave;
	mod.format = fmt_mod;

	sysf_seek(&file->fh, 0, SEEK_SET);

	length = sysf_filelen(&file->fh);
	if (length < 0) return jfauderr_notrecognised;
	if (length > LARGESTMOD) return jfauderr_notrecognised;

	data = malloc(length);
	if (!data) return jfauderr_outofmemory;

	if (sysf_read(&file->fh, data, length) != length) {
		free(data);
		return jfauderr_outofmemory;
	}

	mod.mod = ModPlug_Load(data,length);
	if (!mod.mod) {
		free(data);
		return jfauderr_notrecognised;	// FIXME: this should really check if out of memory or something...
	}

	file->sf = (SoundFile *)malloc(sizeof(ModPlug));
	if (!file->sf) {
		ModPlug_Unload(mod.mod);
		return jfauderr_outofmemory;
	}

	sysf_close(&file->fh);	// we don't need to keep the file open

	memcpy(file->sf, &mod, sizeof(ModPlug));

	return jfauderr_noerror;
}

static jfauderr unload(File *file, char *noclose)
{
	ModPlug *mod = (ModPlug*)file->sf;
	
	testifloaded();

	if (mod->mod) ModPlug_Unload(mod->mod);
	// caller cleans up file->sf
	
	*noclose = 1;
	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file UNUSED, JFAudMetaInfo what UNUSED, char *buf UNUSED, int *len UNUSED)
{
	return jfauderr_notimpl;
}


// WaveformFileReader functions
static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop UNUSED)
{
	ModPlug *mod = (ModPlug *)file->sf;
	int br, nb;
	jfauderr err;

	testifloaded();

	nb = 2 * 2 * n;	// 2bytes * 2channels

	err = pcmbuffer_allocate(buf, n, min(conf.samplerate,44100), 2, 16);
	if (err != jfauderr_noerror) return err;

	br = ModPlug_Read(mod->mod, buf->data, nb);
	buf->numsamples = br / (2*2);

	return jfauderr_noerror;
}

static jfauderr readallsamples(File *file UNUSED, PcmBuffer *buf UNUSED)
{
	// not needed
	return jfauderr_notimpl;
}

static jfauderr getplaytime(File *file UNUSED, float *len)
{
	testifloaded();

	*len = -1.0;	// always stream modules
	
	return jfauderr_noerror;
}

#endif

WaveformFileReader wf_mod = {
	"MOD/XM/IT/S3M/etc",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_mod,
	readsamples,
	readallsamples,
	getplaytime
};

