#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "internal.h"

typedef struct {
	WaveformFile_defn

	unsigned int samplerate;
	unsigned char channels, bitspersample, filler[2];

	unsigned long blocklen;

	unsigned long datastart,datalen,datapos,blocksize;
} Voc;

#ifdef _MSC_VER
#pragma pack(1)
#endif

typedef union {
	struct PACKED {
		unsigned char tc;
		unsigned char compr;
	} block1;
	struct PACKED {
		unsigned short length;
		unsigned char tc;
	} block3;
	struct PACKED {
		unsigned short id;
	} block4;
	struct PACKED {
		unsigned short repeat;
	} block6;
	struct PACKED {
		unsigned short tc;
		unsigned char pack;
		unsigned char mode;
	} block8;
	struct PACKED {
		unsigned long samplerate;
		unsigned char bitspersample;
		unsigned char channels;
		unsigned short format;
		char filler[4];
	} block9;
} VocBlocks;

#ifdef _MSC_VER
#pragma pack()
#endif


// SoundFileReader functions

static jfauderr init(void)
{
	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	char sig[20];
	short offset = 0;
	short version = 0, version2 = 0;
	VocBlocks blk;
	jfauderr err;
	int br;
	
	Voc voc;
	
	memset(&voc, 0, sizeof(voc));
	voc.type = type_wave;
	voc.format = fmt_voc;
	
	if (!file) return jfauderr_genericerror;
	if (file->sf) return jfauderr_genericerror;	// already been identified

	sysf_seek(&file->fh, 0, SEEK_SET);

	if (sysf_read(&file->fh, &sig, 20) != 20)
		return jfauderr_notrecognised;
	if (memcmp(sig, "Creative Voice File\x1a", 20))
		return jfauderr_notrecognised;

	br  = sysf_read(&file->fh, &offset, 2);
	br += sysf_read(&file->fh, &version, 2);
	br += sysf_read(&file->fh, &version2, 2);
	if (br != 6) return jfauderr_filedamaged;

	if (~version + 0x1234 != version2) return jfauderr_filedamaged;

	voc.datastart = offset;

	err = jfauderr_ok;
	while (err == jfauderr_ok) {
		char blocktyp = 0;
		long blocklen = 0;
		
		if (sysf_read(&file->fh, &blocktyp, 1) != 1) { err = jfauderr_filedamaged; break; }
		if (blocktyp == 0) break;	// terminator

		if (sysf_read(&file->fh, &blocklen, 3) != 3) { err = jfauderr_filedamaged; break; }

		switch (blocktyp) {
			case 1:
				if (sysf_read(&file->fh, &blk, 2) != 2) { err = jfauderr_filedamaged; break; }
				blocklen -= 2;
				if (blk.block1.compr != 0) { err = jfauderr_notcompatible; break; }
				if (!voc.samplerate) {
					voc.samplerate = 1000000/(256-blk.block1.tc); // tc=256-(1000000/sr)
					voc.channels = 1;
					voc.bitspersample = 8;
				} else {
					if (voc.samplerate != (unsigned int)(1000000/(256-blk.block1.tc))) {
						err = jfauderr_notcompatible; break;
					}
				}
				voc.datalen += blocklen;
				break;
			case 2:
				if (!voc.samplerate) { err = jfauderr_filedamaged; break; }
				voc.datalen += blocklen;
			case 8:
				if (sysf_read(&file->fh, &blk, 4) != 4) { err = jfauderr_filedamaged; break; }
				blocklen -= 4;
				if (!voc.samplerate) {
					voc.samplerate = 256000000/(65536-blk.block8.tc); // tc=65536-(256000000/sr)
					voc.samplerate >>= blk.block8.mode;
					voc.channels = blk.block8.mode + 1;
					voc.bitspersample = 8;
				} else {
					err = jfauderr_notcompatible; break;
				}
				break;
			case 9:
				if (sysf_read(&file->fh, &blk, 12) != 12) { err = jfauderr_filedamaged; break; }
				blocklen -= 12;
				voc.samplerate = blk.block9.samplerate;
				voc.channels = blk.block9.channels;
				voc.bitspersample = blk.block9.bitspersample;
				voc.datalen += blocklen;
				break;
			default: break;
		}
		if (blocklen > 0) sysf_seek(&file->fh, blocklen, SEEK_CUR);
	}

	if (err != jfauderr_ok) return err;
	
	voc.blocksize = voc.channels * voc.bitspersample / 8;
	sysf_seek(&file->fh, voc.datastart, SEEK_SET);
	
	file->sf = (SoundFile *)malloc(sizeof(Voc));
	if (!file->sf) return jfauderr_outofmemory;	// out of memory

	memcpy(file->sf, &voc, sizeof(Voc));

	return jfauderr_noerror;
}

static jfauderr unload(File *file UNUSED, char *noclose)
{
	// caller cleans up file->sf
	*noclose = 0;

	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file UNUSED, JFAudMetaInfo what UNUSED, char *buf UNUSED, int *len UNUSED)
{
	return jfauderr_notimpl;
}


// WaveformFileReader functions
static jfauderr loadsound(File *file, Voc *voc, PcmBuffer *buf, unsigned long samples, char loop)
{
	int readbytes, justrewound = 0;
	jfauderr err;
	char *rp;

	err = pcmbuffer_allocate(buf, samples, voc->samplerate, voc->channels, voc->bitspersample);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;
	
	rp = (char *)buf->data;
	do {
		if (voc->blocklen == 0) {
			char blocktyp = 0;
			long blocklen = 0;

			if (sysf_read(&file->fh, &blocktyp, 1) != 1) {
				pcmbuffer_free(buf);
				return jfauderr_ioerror;
			}

			if (blocktyp == 0) {
				if (!loop) {
					break;
				} else {
					if (justrewound) loop = 0;
					else {
						sysf_seek(&file->fh, voc->datastart, SEEK_SET);
						voc->datapos = 0;
						justrewound = 1;
					}
				}
			} else {
				if (sysf_read(&file->fh, &blocklen, 3) != 3) {
					pcmbuffer_free(buf);
					return jfauderr_ioerror;
				}
				
				switch (blocktyp) {
					case 1: sysf_seek(&file->fh, 2, SEEK_CUR); voc->blocklen = blocklen - 2; break;
					case 2: voc->blocklen = blocklen; break;
					case 9: sysf_seek(&file->fh, 12, SEEK_CUR); voc->blocklen = blocklen - 12; break;
					default: sysf_seek(&file->fh, blocklen, SEEK_CUR); break;
				}
			}
		} else {
			unsigned long s;

			readbytes = sysf_read(&file->fh, rp, min(samples * voc->blocksize, voc->blocklen));
			if (readbytes < 0) {
				pcmbuffer_free(buf);
				return jfauderr_ioerror;
			} else if (readbytes == 0) {
				pcmbuffer_free(buf);
				return jfauderr_filedamaged;
			}

			s = readbytes / voc->blocksize;
			buf->numsamples += s;

			samples -= s;
			justrewound = 0;

#if MustSwapLittle
			if (voc->bitspersample == 16) {
				unsigned long i;
				unsigned short *p;
				// byte swap shorts
				p = (unsigned short *)rp;
				for (i=readbytes/2; i; i--, p++) {
					*p = LittleShort(*p);
				}
			}
#endif
			voc->datapos += readbytes;
			voc->blocklen -= readbytes;
			rp += readbytes;
		}
	} while (samples > 0);

	if (buf->numsamples > 0)
		debuglog(DEBUGLVL_WORDY,jsprintf("Read %d sample(s)", buf->numsamples));

	return jfauderr_noerror;
}

static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	Voc *voc;

	voc = (Voc *)file->sf;

	if (n <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	return loadsound(file, voc, buf, n, loop);
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	Voc *voc;
	unsigned long samples;

	voc = (Voc *)file->sf;

	samples = voc->datalen / voc->blocksize;
	if (samples == 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	sysf_seek(&file->fh, voc->datastart, SEEK_SET); voc->datapos = 0;
	voc->blocklen = 0;
	
	return loadsound(file, voc, buf, samples, 0);
}

static jfauderr getplaytime(File *file, float *len)
{
	Voc *voc;

	voc = (Voc *)file->sf;

	*len = (float)(voc->datalen / voc->blocksize) / (float)voc->samplerate;
	
	return jfauderr_noerror;
}


WaveformFileReader wf_voc = {
	"Creative VOC",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_voc,
	readsamples,
	readallsamples,
	getplaytime
};

